import pytest
import json

from ..dbc_reader import read_dbc

@pytest.fixture
def dbc():
    with open("dbc/FST09e.dbc") as f:
        return f.read()

@pytest.fixture
def device_config():
    with open("json/device_config.json") as f:
        return json.loads(f.read())

def test_read_dbc():
    read_dbc("dbc/FST09e.dbc", "test.json", "json/device_config.json")








import pytest

from hypothesis import given
import hypothesis.strategies as st

from ...specs import Signal, SignalValueError
from ..test_spec import spec, json

from .test_message import message, device

from ...can import CANMessage


@pytest.fixture(scope="session")
def signal(message):
    return Signal(parent=message, start=0, length=16, scale=1, offset=0)


def test_signal_encode(signal):
    assert signal.encode(10) == 10

def test_signal_encode_scale(signal):
    signal.scale = 0.1
    assert signal.encode(10) == 100

def test_signal_encode_offset(signal):
    signal.scale = 10
    signal.offset = 50
    assert signal.encode(150) == 10
    signal.scale = 1
    signal.offset = 0

@given(s = st.integers(min_value=0))
def test_decode_signal_encode(signal, s):
    try:
        data64 = signal.encode(s)
        return
    except SignalValueError:
        return
    msg = CANMessage(sid=1, dlc=8, data64=data64, timestamp=0)
    assert signal.decode(msg) == s



import pytest

from ...specs import *
from ...can import CANMessage

from ..test_spec import spec, json

@pytest.fixture
def common():
    return Common()

@pytest.fixture
def send_cmd(common):
    msg = Message(parent=common)
    common.msgs["send_cmd"] = msg
    return msg

@pytest.fixture
def cmd(spec):
    devs = list(spec.devices.values())
    return Command(parent=devs[0])

def test_cmd_encode(send_cmd, cmd):
    """test encode of send cmd"""
    args = [1,2,3]
    msg = cmd.encode(1, 10, args)

    assert type(msg) is CANMessage
    datas = msg.get_data16()
    assert datas[1] == args[0]
    assert datas[2] == args[1]
    assert datas[3] == args[2]

def test_cmd_is_cmd(cmd, common):
    args=[1,2,3]
    msg = cmd.encode(1,10,args)
    assert cmd.is_cmd(msg)
    msg = CANMessage(sid=63<<5, dlc=8, data16=[1,2,3,4], timestamp=0)
    assert not cmd.is_cmd(msg)





from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt, Signal

from .undo_redo import UndoRedo, UndoState, UndoUpdate


class FakeParent:
    def __init__(self):
        return

    def setVisible(self, visibility: bool):
        return


class NodeDetails(QWidget):
    """ Node (Device, Message, Signal, Log, Config, Command) interface """

    reload_sig = Signal(str)

    def __init__(self, gui, node, parent=None, details=None, layout=None):
        QWidget.__init__(self)
        self.node = node
        self.gui = gui
        self.parent = parent
        self.atts = []
        self.details = None
        self.Details = details
        self.layout = layout
        self.children = []
        self.details_button = None
        self.reload_sig.connect(self.gui.reload_service)

    def load_atts(self, ui, node):
        self.atts = []
        return

    def connect_atts(self):
        def store(obj, var, att):
            def closure():
                old = getattr(obj,var)
                value = type(old)(att.text())
                setattr(obj, "_" + var, value)

                self.gui.undo_redo.push(UndoUpdate(obj, var, old, att.text()))
                self.gui.reload()
                self.reload_sig.emit("")

            return closure

        for att, var, var_name in self.atts:
            att.editingFinished.connect(store(self.node, var_name, att))

    def save(self):
        for child in self.children:
            child.save()

        for att, var, var_name in self.atts:
            var = att.text()

        self.gui.reload()

    def delete(self):
        r = self.gui.spec.rm_node(self.node)
        self.setVisible(False)
        if self.parent:
            self.parent.setVisible(False)

        if self.details:
            self.details.setVisible(False)
        return

    def reload(self):
        for child in self.children:
            child.reload()

        for att, var, var_name in self.atts:
            att.setText(str(getattr(self.node, var_name)))

    def raise_widget(self, checked):
        self.setVisible(checked)

    def open_details(self, checked):
        if checked == False:
            for child in self.children:
                child.open_details(False)
                if child.details_button is not None:
                    child.details_button.setChecked(False)

        if self.details_button is None:
            return

        if self.details is None:
            self.details = self.Details(self.gui, self.node, self)
            self.layout.addWidget(self.details)
            self.children.append(self.details)

        self.details.setVisible(checked)

    def show(self):
        if self.parent:
            self.parent.setVisible(True)

        if self.details:
            self.details.setVisible(True)
            self.details_button.setFlat(False)

        self.setVisible(True)



from typing import *

from ..can import CANMessage

class Node():
    @property
    def spec(self):
        if type(self.parent) is Spec:
            return self.parent

        if self.parent is None:
            exit()

        return self.parent.get_spec()

    def filter_private(self, d: Dict[str, Any]) -> Dict[str, Any]:
        return {k:v for (k,v) in d.items() if k.startswith('_')}

    def make_private(self, obj, d: Dict[str, Any]) -> Dict[str, Any]:
        new_d = {}
        for key in d.keys():
            T = type(getattr(obj, key))
            new_d["_" + key] = T(d[key])
        return new_d

    def make_public(self, obj, d: Dict[str, Any]) -> Dict[str, Any]:
        return {k[1:] if k.startswith('_') else k:v for (k,v) in d.items()}

    def normalize(self):
        return

class Transmit(Node):
    def encode(self):
        print("encode not implemented")
        assert False

    def decode(self, msg: CANMessage):
        print("decode not implemented")
        assert False

    pass



from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt
from PySide2.QtCore import Signal

from .widgets import *

from .node_details import NodeDetails, FakeParent
from .signal_details import SignalDetails
from .signal_widget import SignalWidget
from .message_details import MessageDetails
from .message_widget import MessageWidget
from .arg_details import ArgDetails
from .cmd_details import CmdDetails
from .cmd_widget import CmdWidget
from .cfg_details import CfgDetails
from .cfg_widget import CfgWidget
from .device_details import DeviceDetails
from .device_widget import DeviceWidget
from .enum_value_details import EnumValueDetails
from .enum_details import EnumDetails
from .log_details import LogDetails
from .log_widget import LogWidget
from .enum_widget import EnumWidget

from pathlib import Path
import json
import yaml
import webbrowser
from copy import deepcopy, copy

from ..version import VERSION
from ..specs import *
from ..validator import validate
from ..config import *
from .message_box import MessageBox

from .undo_redo import UndoRedo, UndoAdd


import requests


def nag_intro():
    def version_comparison(v1, v2):
        v1s = [int(v) for v in v1.split(".")]
        v2s = [int(v) for v in v2.split(".")]

        for v1, v2 in zip(v1s, v2s):
            if v1 > v2:
                return 1
            elif v1 < v2:
                return -1

        return 0

    try:
        r = requests.get("https://pypi.org/pypi/fcp/json")
    except Exception as e:
        return
    j = json.loads(r.text)
    releases = list(j["releases"].keys())
    releases.sort(key=lambda s: [int(u) for u in s.split('.')])
    upstream_version = releases[-1]

    out = ""

    version_comp = version_comparison(upstream_version, VERSION)
    if version_comp == 1:
        out += f"<p><b>FCP v{upstream_version} is available. Go get it:\nsudo pip install fcp=={upstream_version}</b></p>"
    elif version_comp == 0:
        release_url = f"https://joajfreitas.gitlab.io/fcp-core/v{VERSION}.html"
        r = requests.get(release_url)
        if r.ok:
            out += "\n" + r.text
    elif version_comp == -1:
        out += f"Oh! I see you're running a development version. Good luck."

    return out

class MainWindow(QMainWindow):
    def __init__(self):
        QMainWindow.__init__(self)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.history = None

        self.spec = Spec()
        self.children = []

        self.connect_buttons()
        self.config_shortcuts()

        self.log_widget = None
        self.enum_widget = None

        self.filename = None

        self.config, _ = config_session()
        self.recent_files()

        self.undo_redo = UndoRedo()



        MessageBox(
            self,
            QMessageBox.Ok,
            QMessageBox.Information,
            nag_intro()).launch()


    def recent_files(self):
        def loads(path):
            def closure():
                self.load(path)

            return closure

        files = File.recent_files(self.config)
        menu = QMenu("recent_files")

        for file in files:
            action = menu.addAction(file.path)
            action.triggered.connect(loads(file.path))

        self.ui.actionOpen_Recent.setMenu(menu)

    def connect_buttons(self):
        def fcp_help(link):
            webbrowser.open(link)

        self.ui.actionOpen.triggered.connect(self.open_json)
        self.ui.actionSave.triggered.connect(self.save_json)
        self.ui.actionValidate.triggered.connect(self.validate)
        self.ui.action_software10e_help.triggered.connect(
            lambda: fcp_help("https://projectofst.gitlab.io/software10e/docs/fcp/")
        )
        self.ui.action_fcp_help.triggered.connect(
            lambda: fcp_help("https://fcp-core.readthedocs.io/en/latest/")
        )
        self.ui.addButton.clicked.connect(self.add_device)

    def shortcut(self, key_sequence, f):
        shortcutSave = QShortcut(key_sequence, self)
        shortcutSave.setContext(Qt.ApplicationShortcut)
        shortcutSave.activated.connect(f)


    def config_shortcuts(self):
        self.shortcut(QKeySequence(Qt.CTRL + Qt.Key_S), self.save_json)
        self.shortcut(QKeySequence(Qt.CTRL + Qt.Key_O), self.open_json)
        self.shortcut(QKeySequence(Qt.CTRL + Qt.Key_V), self.validate)
        self.shortcut(QKeySequence(Qt.CTRL + Qt.Key_Z), self.undo)
        self.shortcut(QKeySequence(Qt.CTRL + Qt.Key_R), self.redo)

    def save_history(self):
        old_spec = deepcopy(self.spec)
        self.history.insert(0, old_spec)

    def undo(self):
        self.undo_redo.undo()
        self.reload()


    def redo(self):
        self.undo_redo.redo()
        self.reload()

    def validate(self) -> int:
        failed = validate(self.spec)
        failed_count = len([lvl for lvl, msg in failed if lvl == "error"])
        if len(failed) == 0:
            MessageBox(
                self,
                QMessageBox.Ok,
                QMessageBox.Information,
                "Spec passed").launch()
        else:
            if len(failed) > 5:
                failed = sorted(failed, key = lambda x : 0 if x[0] == "error" else 1)
                failed = [(lvl, msg) for lvl, msg in failed]
                failed = [f"{level}: {msg}" for level, msg in failed]
                errors = "\n".join(failed[:5]) + f"\nand {len(failed)-5} more errors..."
            else:
                failed = [f"{level}: {msg}" for level, msg in failed]
                errors = "\n".join(failed)


            MessageBox(
                self,
                QMessageBox.Ok,
                QMessageBox.Warning,
                errors).launch()

        return failed_count

    def add_device(self, device=None, widget=None):
        new_device = device is None or type(device) == bool

        if new_device:
            device = Device(parent=self.spec, msgs={})

        if type(device) is Device:
            r = self.spec.add_device(device)
            #if r == False:
            #    msg = QMessageBox(self)
            #    msg.setStandardButtons(QMessageBox.Ok)
            #    msg.setIcon(QMessageBox.Warning)
            #    msg.setText("Failed to create device")
            #    msg.show()
            #    return

        if widget is None:
            dev_widget = DeviceWidget(
                self, device, details=DeviceDetails, layout=self.ui.deviceDetailsLayout
            )
        else:
            dev_widget = widget

        if new_device:
            undo_action = UndoAdd(device, dev_widget, dev_widget.delete, self.add_device)
            self.undo_redo.push(undo_action)

        self.ui.verticalLayout.addWidget(dev_widget)
        self.children.append(dev_widget)

    def open_json(self):
        filename = QFileDialog.getOpenFileName(
            self, self.tr("Open JSON"), self.tr("JSON (*.json)")
        )
        self.load_json(filename[0])
        self.history = []
        self.history.append()

    def save_json(self):
        for child in self.children:
            child.save()

        l = self.validate()
        if l > 0:
            return

        if self.filename:
            filename = self.filename
        else:
            try:
                filename = QFileDialog.getSaveFileName(
                    self, self.tr("Open JSON"), str(self.filename)
                )
                filename = filename[0]
            except Exception as e:
                msg = QMessageBox(self)
                msg.setStandardButtons(QMessageBox.Ok)
                msg.setIcon(QMessageBox.Warning)
                msg.setText(f"{filename} is not a valid filename")
                msg.show()
                return

        MessageBox(
            self,
            QMessageBox.Ok,
            QMessageBox.Information,
            "Saved").launch()

        with open(filename, "w") as f:
            j = self.spec.compile()
            f.write(json.dumps(j, indent=4))

    def load(self, filename):
        File.access(self.config, filename)
        with open(filename) as f:
            r = f.read()
            j = json.loads(r)

        self.spec.decompile(j)

        self.reload_spec()

        self.filename = Path(filename)
        self.history = [self.spec]

    def load_json(self, filename):
        if filename == "":
            msg = QMessageBox(self)
            msg.setStandardButtons(QMessageBox.Ok)
            msg.setIcon(QMessageBox.Warning)
            msg.setText(f"'{filename}' is not a valid filename")
            msg.show()
            return


        self.filename = filename

        File.access(self.config, filename)
        self.load(filename)


    def reload_spec(self):
        for device in sorted(self.spec.devices.values(), key=lambda x: x.id):
            self.add_device(device)

        self.add_device(self.spec.common)

        self.log_widget = LogWidget(self, self.spec)
        self.log_widget.setVisible(True)
        self.ui.logDetailsLayout.addWidget(self.log_widget)
        self.children.append(self.log_widget)
        self.enum_widget = EnumWidget(self, self.spec)
        self.enum_widget.setVisible(True)
        self.ui.enumDetailsLayout.addWidget(self.enum_widget)
        self.children.append(self.enum_widget)

    def close_spec(self):
        return

    def reload(self):
        #print("reload:", self.history)
        self.save_history()
        for node in self.children:
            node.reload()

        self.spec.normalize()

    def reload_service(self, path):
        self.reload()



from typing import *
import datetime

from ..can import CANMessage

from .node import Node

class SignalValueError(Exception):
    pass

class Signal(Node):
    """
    Signal node. Represents a CAN signal, similar to a DBC signal.

    :param name: Name of the Signal.
    :param start: Start bit
    :param length: Signal bit size.
    :param scale: Scaling applied to the signal's data.
    :param offset: Offset applied to the signal's data.
    :param unit: Unit of the Signal after applying scaling and offset.
    :param comment: Description of the Signal.
    :param min_value: Minimum value allowed to the Signal's data.
    :param max_value: Maximum value allowed to the Signal's data.
    :param type: Type of the Signal's data.
    :param mux: Name of the mux Signal. None if the Signal doesn't belong to a multiplexed Message.
    :param mux_count: Number of signals that the mux can reference for this Muxed signal.


    """

    def __init__(
        self,
        parent: "Message" = None,
        name: str = "",
        start: int = 0,
        length: int = 0,
        scale: float = 1.0,
        offset: float = 0.0,
        unit: str = "",
        comment: str = "",
        min_value: float = 0.0,
        max_value: float = 0.0,
        type: str = "unsigned",
        byte_order: str = "little_endian",
        mux: str = "",
        mux_count: int = 1,
        alias: str = "",
    ):

        assert parent is not None
        self.parent = parent
        m = max(
            [
                int(sig.name[3:])
                for sig in self.parent.signals.values()
                if sig.name.startswith("sig")
            ]
            + [0]
        )
        self._name = name or ("sig" + str(m + 1))
        self._start = start
        self._length = length
        self._scale = scale
        self._offset = offset
        self._unit = unit
        self._comment = comment
        self._min_value = min_value
        self._max_value = max_value
        self._type = type
        self._byte_order = byte_order
        self._mux = mux
        self._mux_count = mux_count
        self._alias = alias

        self.creation_date = datetime.datetime.now()

    @property
    def name(self) -> str:
        return self._name

    @property
    def start(self) -> int:
        return self._start

    @property
    def length(self) -> int:
        return self._length

    @property
    def scale(self) -> float:
        return self._scale

    @property
    def offset(self) -> float:
        return self._offset

    @property
    def unit(self) -> str:
        return self._unit

    @property
    def comment(self) -> str:
        return self._comment

    @property
    def min_value(self) -> float:
        return self._min_value

    @property
    def max_value(self) -> float:
        return self._max_value

    @property
    def type(self) -> str:
        return self._type

    @property
    def byte_order(self) -> str:
        return self._byte_order

    @property
    def mux(self) -> str:
        return self._mux

    @property
    def mux_count(self) -> int:
        return self._mux_count

    @property
    def alias(self) -> str:
        return self._alias

    @name.setter
    def name(self, name: str) -> None:
        try:
            self._name = name
        except Exception as e:
            return
    @start.setter
    def start(self, start: int) -> None:
        try:
            self._start = int(start)
        except Exception as e:
            return

    @length.setter
    def length(self, length: int) -> None:
        try:
            self._length = int(length)
        except Exception as e:
            return

    @scale.setter
    def scale(self, scale: float) -> None:
        try:
            self._scale = float(scale)
        except Exception as e:
            return

    @offset.setter
    def offset(self, offset: float) -> None:
        try:
            self._offset = float(offset)
        except Exception as e:
            return

    @unit.setter
    def unit(self, unit: str) -> None:
        try:
            self._unit = unit
        except Exception as e:
            return

    @comment.setter
    def comment(self, comment: str) -> None:
        try:
            self._comment = comment
        except Exception as e:
            return

    @min_value.setter
    def min_value(self, min_value: float) -> None:
        try:
            self._min_value = float(min_value)
        except Exception as e:
            return

    @max_value.setter
    def max_value(self, max_value: float) -> None:
        try:
            self._max_value = float(max_value)
        except Exception as e:
            return

    @type.setter
    def type(self, type: str) -> None:
        try:
            self._type = type
        except Exception as e:
            return

    @byte_order.setter
    def byte_order(self, byte_order: str) -> None:
        try:
            self._byte_order = byte_order
        except Exception as e:
            return

    @mux.setter
    def mux(self, mux: str) -> None:
        try:
            self._mux = mux
        except Exception as e:
            return

    @mux_count.setter
    def mux_count(self, mux_count: int) -> None:
        try:
            self._mux_count = int(mux_count)
        except Exception as e:
            return

    @alias.setter
    def alias(self, alias: str) -> None:
        try:
            self._alias = alias
        except Exception as e:
            return

    def compile(self) -> Dict[str, Any]:
        """Transform python class node to its dictionary representation.

        :return: A dictionary containing the node parameters
        """

        return self.make_public(self, self.filter_private(self.__dict__))

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """
        for k,v in self.make_private(self, d).items():
            self.__setattr__(k,v)

    def bitmask(self, n):
        return 2 ** n - 1

    def encode(self, value) -> int:
        if value>2**self.length:
            raise SignalValueError()
        value = (value-self.offset) / self.scale
        return (int(value) & self.bitmask(self.length)) << self.start

    def decode(self, msg: CANMessage):
        data64 = msg.get_data64()
        value = (data64 >> self.start) & self.bitmask(self.length)
        if self.type == "signed" and value >> (self.length - 1) == 1:
            value = -((value ^ self.bitmask(self.length)) + 1);
        return self.scale * value - self.offset

    def __hash__(self):
        return hash((self.name, self.start, self.length, self.creation_date))

    def __repr__(self):
        return f"<Signal name={self.name} start={self.start} end={self.length} scale={self.scale} offset={self.offset}>"



from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .widgets.enumdetails import Ui_EnumDetails
from .node_details import NodeDetails, FakeParent
from .enum_value_details import EnumValueDetails
from ..specs import EnumValue


class EnumDetails(NodeDetails):
    def load_atts(self, ui, node):
        self.atts = [
            (ui.nameEdit, node.name, "name"),
        ]

    def __init__(self, gui, node, parent):
        NodeDetails.__init__(self, gui, node, parent)

        self.ui = Ui_EnumDetails()
        self.ui.setupUi(self)

        self.load_atts(self.ui, node)
        self.connect_atts()

        self.ui.enumDeleteButton.clicked.connect(self.delete)

        self.reload()
        self.setVisible(True)

        self.children = []

        self.ui.valueAddButton.clicked.connect(self.add_value)

        for enum_value in node.enumeration.values():
            self.add_value(enum_value)

    def add_value(self, enum_value=None):
        if enum_value == None or enum_value == False:
            enum_value = EnumValue(self.node)
            self.node.enumeration[enum_value.name] = enum_value

        enum_widget = EnumValueDetails(self.gui, enum_value, FakeParent())
        self.children.append(enum_widget)
        self.ui.EnumValueContents.addWidget(enum_widget)

    def save(self):
        for child in self.children:
            child.save()

        self.gui.reload()



import copy

from PySide2.QtCore import Signal
from PySide2.QtWidgets import *

class UndoState():
    def __init__(self):
        pass

    def undo(self):
        print("UndoState should not be used directly")
        exit()


class UndoUpdate(UndoState):
    def __init__(self, obj, var, old, new):
        self.obj = obj
        self.var = var
        self.old = old
        self.new = new

        if old == new:
            return None

    def undo(self):
        setattr(self.obj, self.var, self.old)

    def redo(self):
        setattr(self.obj, self.var, self.new)

    def __repr__(self):
        return f"{self.var} {self.old} {self.new}"

class UndoAdd(UndoState):
    #delete_sig = Signal()
    def __init__(self, obj, holder, remove, add):
        self.obj = obj
        self.holder = holder
        self.remove = remove
        self.add = add

    def undo(self):
        print("undo")
        self.obj = copy.copy(self.obj)
        #print("remove from spec", self.obj.spec.rm_node(self.obj))
        self.remove()

    def redo(self):
        self.holder.node = self.obj
        self.holder.show()

    def __repr__(self):
        return f"{self.obj}"

class UndoRedo():
    def __init__(self):
        self.undo_stack = []
        self.redo_stack = []
        return

    def push(self, state):
        print("push", state)
        if state is None:
            return
        self.undo_stack.append(state)
        self.redo_stack = []

    def undo(self):
        if len(self.undo_stack) == 0:
            print("Undo not possible")
            return

        state = self.undo_stack.pop()
        state.undo()
        self.redo_stack.append(state)

    def redo(self):
        if len(self.redo_stack) == 0:
            print("Redo not possible")
            return

        state = self.redo_stack.pop()
        state.redo()
        self.undo_stack.append(state)



from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

class MessageBox():
    def __init__(self, parent, buttons, icon, text):
        self.msg = QMessageBox(parent)
        self.msg.setStandardButtons(buttons)
        self.msg.setIcon(icon)
        self.msg.setText(text)

    def launch(self):
        self.msg.show()




# This Python file uses the following encoding: utf-8
import sys

from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .gui_lib.mainwindow import MainWindow


def gui(file):
    app = QApplication([])
    window = MainWindow()
    if file != None and file != "":
        window.load_json(file)
    window.show()
    sys.exit(app.exec_())



""" Main """

import sys
import logging
import json
import subprocess
import os
from pathlib import Path

import click

from .dbc_reader import read_dbc
from .dbc_writer import write_dbc
from .c_generator import c_gen
from .validator import validate, format_error
from .specs import Spec
from .docs import generate_docs
from .version import VERSION
from .idl import spec_to_fcp_v2, fcp_v2_from_file, fcp_v2


def setup_logging() -> logging.Logger:
    """Setup Logger.

    :return: logger object.
    """

    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    handler = logging.FileHandler("fcp.log")
    handler.setLevel(logging.DEBUG)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def report_validate(failed, force):
    error = 0
    if len(failed):
        for level, msg in failed:
            print(format_error(level, msg))

            if level == "error":
                error += 1

        if error > 0:
            if not force:
                print("Too many error won't continue")
                return True

    return False


def get_spec(json_file: str, force: bool = False) -> Spec:
    """Create Spec from json file path.
    :param json_file: path to the json file.
    :param logger: logger.
    :return: Spec.
    """

    spec = Spec()

    path = Path(json_file)
    if path.suffix == ".json":
        with open(json_file) as f:
            j = json.loads(f.read())
    elif path.suffix == ".fcp":
        j = fcp_v2_from_file(json_file)

    spec.decompile(j)
    failed = validate(spec)
    if report_validate(failed, force):
        exit()

    return spec


@click.command(name="read-dbc")
@click.argument("dbc")
@click.argument("json_file")
@click.argument("device_config")
def read_dbc_cmd(dbc: str, json_file: str, device_config: str):
    """Transform a DBC file into FCP json.
    :param dbc: dbc file path.
    :param json_file: FCP json file path.
    :param device_config: json with device name configuration.

    device_config example:

    ```json
    {
        "0": "amk_0",
        "1": "amk_1",
        "2": "amk_2",
        "3": "amk_3",
        "8": "dcu",
        "9": "te",
        "10": "dash",
        "13": "arm",
        "14": "bms_master",
        "15": "bms_verbose",
        "16": "iib",
        "20": "ahrsf",
        "21": "ahrsr",
        "22": "gpsf",
        "23": "gpsr",
        "29": "isa"
    }```
    """
    logger = setup_logging()
    read_dbc(dbc, json_file, device_config, logger)


@click.command(name="write-dbc")
@click.argument("json_file")
@click.argument("dbc")
def write_dbc_cmd(json_file: str, dbc: str):
    """Transform FCP json file into a DBC
    :param json_file: FCP json file path.
    :param dbc: dbc file path.
    """

    spec = get_spec(json_file)
    write_dbc(spec, dbc)


@click.command(name="c_gen")
@click.argument("templates")
@click.argument("output")
@click.argument("json_file")
@click.argument("skel")
@click.option("--noformat", is_flag=True, default=False)
@click.option("--force", is_flag=True, default=False)
def c_gen_cmd(templates: str, output: str, json_file: str, skel: str, noformat: bool, force: bool):
    """Transform FCP json into a C library.
    :param templates: jinja template directory.
    :param output: output directory.
    :param json_file: FCP json file path.
    :param skel: C skeleton directory.
    """
    logger = setup_logging()

    spec = get_spec(json_file, force)
    c_gen(spec, templates, output, skel, logger)

    if noformat == False:
        subprocess.run(
            "clang-format -i -style=Webkit *.c *.h",
            cwd=os.path.abspath(output),
            shell=True,
        )
        logger.info("clang-format clean up ✅")


@click.command(name="init")
@click.argument("json_file")
def init_cmd(json_file: str):
    """Create a basic FCP json file.
    :param json_file: FCP json file path.
    """

    logger = setup_logging()
    # init(json_file, logger)


@click.command(name="validate")
@click.argument("json_file")
def validate_cmd(json_file: str):
    """Verify correctness of FCP json file.
    :param json_file: FCP json file path.
    """

    logger = setup_logging()
    spec = get_spec(json_file)

    failed = validate(spec)
    if len(failed) == 0:
        print("✓ JSON validated")
    else:
        print("❌")
        for level, msg in failed:
            print(format_error(level, msg))


@click.command(name="gui")
@click.argument("file", required=False)
def gui_cmd(file: str):
    """Launch FCP json editor GUI.
    :param file: Optional FCP json file path
    """

    from .gui import gui

    logger = setup_logging()
    gui(file)


@click.command(name="gui2")
@click.argument("file", type=Path, required=False)
def gui_cmd2(file: Path):
    """Launch FCP json editor GUI.
    :param file: Optional FCP json file path
    """

    from .gui2 import gui2

    gui2(file)


@click.command()
@click.argument("json_file")
def dump_dev_list(json_file):
    spec = get_spec(json_file)

    for dev in spec.devices.keys():
        print(dev)


@click.command()
@click.argument("json_file")
@click.argument("device", required=False)
def dump_msg_list(json_file, device):
    spec = get_spec(json_file)

    for dev_name, dev in spec.devices.items():
        if device != None and device != dev_name:
            continue

        for msg in dev.msgs.keys():
            print(msg)


@click.command()
@click.argument("json_file")
@click.argument("device", required=False)
def dump_cfg_list(json_file, device):
    spec = get_spec(json_file)

    for dev_name, dev in spec.devices.items():
        if device != None and device != dev_name:
            continue

        for cfg in dev.cfgs.keys():
            print(cfg)


@click.command()
@click.argument("json_file")
@click.argument("device", required=False)
def dump_cmd_list(json_file, device):
    spec = get_spec(json_file)

    for dev_name, dev in spec.devices.items():
        if device != None and device != dev_name:
            continue

        for cmd in dev.cmds.keys():
            print(cmd)


@click.command()
@click.argument("json_file")
def dump_log_list(json_file):
    spec = get_spec(json_file)

    for log in spec.logs.keys():
        print(log)


@click.command()
@click.argument("json_file")
@click.argument("message", required=False)
def dump_signal_list(json_file, message):
    spec = get_spec(json_file)

    for dev in spec.devices.values():
        for msg in dev.msgs.values():
            if message != None and msg.name != message:
                continue
            for signal in msg.signals.values():
                print(signal)


@click.command()
@click.argument("json_file")
@click.argument("message", required=False)
def dump_signal_names(json_file, message):
    spec = get_spec(json_file)

    for dev in spec.devices.values():
        for msg in dev.msgs.values():
            if message != None and msg.name != message:
                continue
            for signal in msg.signals.values():
                print(signal.name)


@click.command()
@click.argument("json_file")
@click.argument("log")
def print_log(json_file, log):
    spec = get_spec(json_file)

    print(spec.get_log(log))


@click.command()
@click.argument("json_file")
@click.argument("dev")
def print_dev(json_file, dev):
    spec = get_spec(json_file)

    print(spec.get_device(dev))


@click.command()
@click.argument("json_file")
@click.argument("message")
def print_msg(json_file, message):
    spec = get_spec(json_file)

    for dev in spec.devices.values():
        for name, msg in dev.msgs.items():
            if name == message:
                print(dev.name)
                print(msg)


@click.command()
@click.argument("json_file")
@click.argument("signal")
def print_signal(json_file, signal):
    spec = get_spec(json_file)

    for dev in spec.devices.values():
        for msg in dev.msgs.values():
            for name, sig in msg.signals.items():
                if name == signal:
                    print(dev.name, msg.name)
                    print(sig)


@click.command("docs")
@click.argument("json_file")
@click.argument("out")
@click.argument("link_location", required=False)
def docs(json_file: str, out: str, link_location: str):
    """Generate FCP documentation.
    :param json_file: FCP json file path.
    :param out: output directory.
    """

    if link_location == None:
        link_location = "."

    logger = setup_logging()

    spec = get_spec(json_file)
    generate_docs(spec, out, link_location, logger)

@click.command("fix")
@click.argument("src")
@click.argument("dst")
def fix(src: str, dst: str):
    print("Warning: This operation may lose some data")
    spec = Spec()
    with open(src) as f:
        j = json.loads(f.read())

    spec.decompile(j)

    for dev in spec.devices.values():
        for msg in dev.msgs.values():
            for sig in msg.signals.values():
                if sig.mux_count == 0:
                    sig.mux_count = 1


    d = spec.compile()
    with open(dst, "w") as f:
        f.write(json.dumps(d, indent=4))

@click.command("json_to_fcp2")
@click.argument("json_file")
@click.argument("output")
def json_to_fcp2(json_file: str, output: str):
    spec = get_spec(json_file)
    v2 = spec_to_fcp_v2(spec)

    with open(output, "w") as f:
        f.write(v2)



@click.command("read_fcp2")
@click.argument("fcp_v2_file")
def read_fcp2(fcp_v2_file):
    with open(fcp_v2_file) as f:
        v2 = fcp_v2(f.read())

    spec = Spec()
    spec.decompile(v2)


@click.command("read_fcp")
@click.argument("json_file")
def read_fcp(json_file: str):
    logger = setup_logging()

    spec = get_spec(json_file)
    print(spec)



@click.group(invoke_without_command=True)
@click.option("--version", is_flag=True, default=False)
def main(version):
    """CLI utility for managment of FCP JSON files. """

    if len(sys.argv) == 1:
        print("fcp cli util.\nVersion:", VERSION, "\nFor usage see fcp --help")
    if version:
        click.echo(VERSION)
    pass


main.add_command(read_dbc_cmd)
main.add_command(write_dbc_cmd)
main.add_command(c_gen_cmd)
main.add_command(init_cmd)
main.add_command(validate_cmd)
main.add_command(gui_cmd)
#main.add_command(dump_dev_list)
#main.add_command(dump_msg_list)
#main.add_command(dump_cfg_list)
#main.add_command(dump_cmd_list)
#main.add_command(dump_log_list)
#main.add_command(dump_signal_list)
#main.add_command(dump_signal_names)
#main.add_command(print_log)
#main.add_command(print_dev)
#main.add_command(print_msg)
#main.add_command(print_signal)
main.add_command(docs)
main.add_command(fix)
main.add_command(json_to_fcp2)
main.add_command(read_fcp2)
main.add_command(read_fcp)

if __name__ == "__main__":
    main()



from typing import *
import datetime

from .node import Node, Transmit
from .message import Message
from .config import Config
from .cmd import Command

class Common(Node):
    def __init__(
        self,
        parent: "Spec" = None,
        name: str = "common",
        id: int = 0,
        msgs: Dict[str, Message] = None,
        cfgs: Dict[str, Config] = None,
        cmds: Dict[str, Command] = None,
    ):
        self.parent = parent
        self._name = name
        self._id = id
        self.msgs = {} if msgs is None else msgs
        self.cfgs = {} if cfgs is None else cfgs
        self.cmds = {} if cmds is None else cmds

        self.creation_date = datetime.datetime.now()

    @property
    def name(self) -> str:
        return self._name

    @property
    def id(self) -> int:
        return self._id

    @name.setter
    def name(self, name: str) -> None:
        try:
            self._name = name
        except Exception as e:
            return

    @id.setter
    def id(self, id: int) -> None:
        try:
            self._id = int(id)
        except Exception as e:
            return


    def add_msg(self, msg: Message) -> bool:
        if msg == None:
            return False

        if msg.name in self.msgs.keys():
            return False

        self.msgs[msg.name] = msg
        return True

    def get_msg(self, name: str) -> Message:
        for msg_name, msg in self.msgs.items():
            if msg_name == name:
                return msg


    def compile(self) -> Dict[str, Any]:
        """Transform python class node to its dictionary representation.

        :return: A dictionary containing the node parameters
        """

        msgs = {}
        for msg_k, msg_v in self.msgs.items():
            msgs[msg_k] = msg_v.compile()

        att = self.make_public(self, self.filter_private(self.__dict__))
        att["msgs"] = msgs
        return att

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """
        msgs = d["msgs"]
        del d["msgs"]

        self.__dict__.update(self.make_private(self, d))

        for key, value in msgs.items():
            msg = Message(self)
            msg.decompile(value)
            self.msgs[key] = msg

    def __hash__(self):
        return hash((self.name, self.id, self.creation_date))

    def __repr__(self):
        return f"<Common name={self.name} id={self.id}>"



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'signalwidget.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_SignalWidget(object):
    def setupUi(self, SignalWidget):
        if not SignalWidget.objectName():
            SignalWidget.setObjectName(u"SignalWidget")
        SignalWidget.resize(348, 77)
        self.horizontalLayout = QHBoxLayout(SignalWidget)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.groupBox = QGroupBox(SignalWidget)
        self.groupBox.setObjectName(u"groupBox")
        self.horizontalLayout_2 = QHBoxLayout(self.groupBox)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.name = QLabel(self.groupBox)
        self.name.setObjectName(u"name")

        self.verticalLayout_2.addWidget(self.name)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.start = QLabel(self.groupBox)
        self.start.setObjectName(u"start")

        self.horizontalLayout_4.addWidget(self.start)

        self.length = QLabel(self.groupBox)
        self.length.setObjectName(u"length")

        self.horizontalLayout_4.addWidget(self.length)


        self.verticalLayout_2.addLayout(self.horizontalLayout_4)


        self.horizontalLayout_2.addLayout(self.verticalLayout_2)

        self.deleteButton = QPushButton(self.groupBox)
        self.deleteButton.setObjectName(u"deleteButton")
        self.deleteButton.setMaximumSize(QSize(25, 25))

        self.horizontalLayout_2.addWidget(self.deleteButton)

        self.signalDetailsButton = QPushButton(self.groupBox)
        self.signalDetailsButton.setObjectName(u"signalDetailsButton")
        self.signalDetailsButton.setMaximumSize(QSize(60, 25))
        self.signalDetailsButton.setCheckable(True)

        self.horizontalLayout_2.addWidget(self.signalDetailsButton)


        self.horizontalLayout.addWidget(self.groupBox)


        self.retranslateUi(SignalWidget)

        QMetaObject.connectSlotsByName(SignalWidget)
    # setupUi

    def retranslateUi(self, SignalWidget):
        SignalWidget.setWindowTitle(QCoreApplication.translate("SignalWidget", u"SignalWidget", None))
        self.groupBox.setTitle("")
        self.name.setText("")
        self.start.setText("")
        self.length.setText("")
        self.deleteButton.setText(QCoreApplication.translate("SignalWidget", u"\u2717", None))
        self.signalDetailsButton.setText(QCoreApplication.translate("SignalWidget", u"...", None))
    # retranslateUi




from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .widgets.logdetails import Ui_LogDetails
from .node_details import NodeDetails, FakeParent


class LogDetails(NodeDetails):
    def load_atts(self, ui, node):
        self.atts = [
            (ui.idEdit, node.id, "id"),
            (ui.nameEdit, node.name, "name"),
            (ui.n_argsEdit, node.n_args, "n_args"),
            (ui.commentEdit, node.comment, "comment"),
            (ui.stringEdit, node.string, "string"),
        ]

    def __init__(self, gui, node, parent):
        NodeDetails.__init__(self, gui, node, parent)

        self.ui = Ui_LogDetails()
        self.ui.setupUi(self)

        self.load_atts(self.ui, node)
        self.connect_atts()

        self.ui.logDeleteButton.clicked.connect(self.delete)

        self.reload()
        self.setVisible(True)

        self.children = []



from typing import *

from appdirs import *
from pathlib import Path
import sys
import hjson

from hashlib import sha1

from sqlalchemy import Column, ForeignKey, Integer, Float, String, func, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session

from .dirs import get_config_dir


SpecBase = declarative_base()


class Signal(SpecBase):
    __tablename__ = "signals"
    _parent = Column(String, ForeignKey("msgs.name"), primary_key=True)
    name = Column(String, primary_key=True, default="signal")
    start = Column(Integer, default=0)
    length = Column(Integer, default=0)
    scale = Column(Float, default=1.0)
    offset = Column(Float, default=0.0)
    unit = Column(String, default="")
    comment = Column(String, default="")
    min_value = Column(Integer, default=0)
    max_value = Column(Integer, default=0)
    type = Column(String, default="unsigned")
    byte_order = Column(String, default="little_endian")
    mux = Column(String, default="")
    mux_count = Column(Integer, default=1)
    alias = Column(String, default="")

    def from_dict(self, fcp, msg_name):
        self._parent = msg_name
        self.__dict__.update(fcp)
        # self.name = fcp["name"]
        # self.start = fcp["start"]
        # self.length = fcp["length"]
        # self.scale = fcp["scale"]
        # self.offset = fcp["offset"]
        # self.unit = fcp["unit"]
        # self.comment = fcp["comment"]
        # self.min_value = fcp["min_value"]
        # self.max_value = fcp["max_value"]
        # self.type = fcp["type"]
        # self.byte_order = fcp["byte_order"]
        # self.mux = fcp["mux"]
        # self.mux_count = fcp["mux_count"]
        # self.alias = fcp["alias"]

    def to_dict(self) -> Dict[Any, Any]:
        return {k: v for (k, v) in self.__dict__.items() if not k.startswith("_")}


class Message(SpecBase):
    __tablename__ = "msgs"
    _parent = Column(String, ForeignKey("devs.name"), primary_key=True)
    name = Column(String, primary_key=True, default="msg")
    id = Column(Integer, default=0)
    dlc = Column(Integer, default=0)
    frequency = Column(Integer, default=1.0)
    description = Column(String, default=0.0)

    def from_dict(self, fcp, dev_name):
        self._parent = dev_name
        self.name = fcp["name"]
        self.id = fcp["id"]
        self.dlc = fcp["dlc"]
        self.frequency = fcp["frequency"]
        self.description = fcp["description"]

    def to_dict(self) -> Dict[Any, Any]:
        return {k: v for (k, v) in self.__dict__.items() if not k.startswith("_")}


class Device(SpecBase):
    __tablename__ = "devs"
    name = Column(String, primary_key=True, default="dev")
    id = Column(Integer, default=0)

    def from_dict(self, fcp):
        self.name = fcp["name"]
        self.id = fcp["id"]

    def to_dict(self) -> Dict[Any, Any]:
        return {k: v for (k, v) in self.__dict__.items() if not k.startswith("_")}

    def __repr__(self):
        return f"<Device name={self.name} id={self.id}>"


class Log(SpecBase):
    __tablename__ = "logs"
    name = Column(String, primary_key=True, default="log")
    id = Column(Integer, default=0)
    n_args = Column(Integer, default=0)
    comment = Column(String, default="")
    string = Column(String, default="Undefined String")

    def from_dict(self, fcp):
        self.name = fcp["name"]
        self.id = fcp["id"]
        self.n_args = fcp["n_args"]
        self.comment = fcp["comment"]
        self.string = fcp["string"]

    def to_dict(self) -> Dict[Any, Any]:
        return {k: v for (k, v) in self.__dict__.items() if not k.startswith("_")}


class Command(SpecBase):
    __tablename__ = "cmds"
    parent = Column(String, ForeignKey("devs.name"), primary_key=True)
    name = Column(String, primary_key=True, default="cmd")
    id = Column(Integer, default=0)
    n_args = Column(Integer, default=0)
    comment = Column(String, default="")

    def from_dict(self, fcp, parent):
        self.parent = parent
        self.name = fcp["name"]
        self.id = fcp["id"]
        self.n_args = fcp["n_args"]
        self.comment = fcp["comment"]


class Config(SpecBase):
    __tablename__ = "cfgs"
    parent = Column(String, ForeignKey("devs.name"), primary_key=True)
    name = Column(String, primary_key=True, default="cfg")
    id = Column(Integer, default=0)
    comment = Column(String, default="")

    def from_dict(self, fcp, parent):
        self.parent = parent
        self.name = fcp["name"]
        self.id = fcp["id"]
        self.comment = fcp["comment"]


def json_to_sql(session: Session, j: Dict[Any, Any]):
    for log in j["logs"].values():
        _log = Log()
        _log.from_dict(log)
        session.add(_log)
    for dev in j["devices"].values():
        device = Device()
        device.from_dict(dev)
        session.add(device)
        for msg in dev["msgs"].values():
            message = Message()
            message.from_dict(msg, dev["name"])
            session.add(message)
            for sig in msg["signals"].values():
                signal = Signal()
                signal.from_dict(sig, msg["name"])
                session.add(signal)

        for cmd in dev["cmds"].values():
            command = Command()
            command.from_dict(cmd, dev["name"])
            session.add(command)

        for cfg in dev["cfgs"].values():
            config = Config()
            config.from_dict(cfg, dev["name"])
            session.add(config)

    session.commit()


def sql_to_json(session: Session) -> Dict[Any, Any]:
    signals = session.query(Signal).all()
    devs = session.query(Device).all()

    json = {"devices": {}}
    json["version"] = 0.3

    for dev in devs:
        json["devices"][dev.name] = dev.to_dict()
        msgs = session.query(Message).filter(Message._parent == dev.name).all()
        json["devices"][dev.name]["msgs"] = {}
        for msg in msgs:
            json["devices"][dev.name]["msgs"][msg.name] = msg.to_dict()
            sigs = session.query(Signal).filter(Signal._parent == msg.name).all()
            json["devices"][dev.name]["msgs"][msg.name]["signals"] = {}
            for sig in sigs:
                json["devices"][dev.name]["msgs"][msg.name]["signals"][
                    sig.name
                ] = sig.to_dict()

    json["logs"] = {}
    logs = session.query(Log).all()
    for log in logs:
        json["logs"][log.name] = log.to_dict()

    return json


def create_session(db_path: Path, base) -> Session:
    engine = create_engine("sqlite:///" + str(db_path))
    base.metadata.create_all(engine)
    base.metadata.bind = engine
    DBSession = sessionmaker(bind=engine)
    session: Session = DBSession()

    return session, engine


def init_session(file_path: Path) -> Session:
    print("file_path:", file_path)
    with file_path.open() as f:
        j = hjson.loads(f.read())

    abs = bytes(str(file_path.absolute()), "utf-8")
    db_path = Path(sha1(abs).hexdigest()[:16] + "-" + file_path.name)
    db_path: Path = Path(get_config_dir()) / db_path

    print("db_path:", db_path)

    if db_path.is_file():
        db_path.unlink()

    session, engine = create_session(db_path, SpecBase)
    json_to_sql(session, j)

    connection = engine.connect()
    # connection.execute()

    return session, engine


def spec_session(file_paht: Path) -> Session:
    db_path = Path(get_config_dir()) / hash(file_path)
    session, _ = create_session(db_path, SpecBase)

    return session


if __name__ == "__main__":
    with open(sys.argv[1]) as f:
        j = hjson.loads(f.read())

    session, _ = create_session("db", SpecBase)
    json_to_sql(session, j)



from typing import *
import datetime

from .node import Node, Transmit
from .arg import Argument
from ..can import CANMessage

class Command(Transmit):
    """Command node. Represents a Command.

    :param name: Name of the Command.
    :param n_args: Number of arguments in the Command.
    :param comment: description of the Command.
    :param id: Command identifier.
    :param args: Dictionary containing the Command's input Arguments.
    :param rets: Dictionary containing the Command's output Arguments.
    """

    def __init__(
        self,
        parent: "Device" = None,
        name: str = "",
        n_args: int = 3,
        comment: str = "",
        id: int = 0,
        args: Dict[str, Argument] = None,
        rets: Dict[str, Argument] = None,
    ):

        self.parent = parent
        assert self.parent is not None

        c = max([int(cmd.id) for cmd in self.parent.cmds.values()] + [0]) + 1

        self._name = name
        self._n_args = n_args
        self._comment = comment
        self._id = int(id)
        self.args = {} if args == None else args
        self.rets = {} if rets == None else rets

        self.creation_date = datetime.datetime.now()

    @property
    def name(self) -> str:
        return self._name

    @property
    def n_args(self) -> int:
        return int(self._n_args)

    @property
    def comment(self) -> str:
        return self._comment

    @property
    def id(self) -> int:
        return int(self._id)

    @name.setter
    def name(self, name: str) -> None:
        try:
            self._name = name
        except Exception as e:
            return

    @n_args.setter
    def n_args(self, n_args: int) -> None:
        try:
            self._n_args = int(n_args)
        except Exception as e:
            return

    @comment.setter
    def comment(self, comment: str) -> None:
        try:
            self._comment = comment
        except Exception as e:
            return

    @id.setter
    def id(self, id: int) -> None:
        try:
            self._id = int(id)
        except Exception as e:
            return

    def compile(self) -> Dict[str, Any]:
        """Transform python class node to its dictionary representation.

        :return: A dictionary containing the node parameters
        """

        args = {}
        rets = {}

        for k, v in self.args.items():
            args[k] = v.compile()

        for k, v in self.rets.items():
            rets[k] = v.compile()

        att = self.make_public(self, self.filter_private(self.__dict__))
        att["args"] = args
        att["rets"] = rets

        return att

    def add_arg(self, arg: Argument) -> None:
        """Add a input Argument to Command.

        :param arg: Argument to be added
        :return: Operation success status: True - Success, False - Failure
        """
        self.args[arg.name] = arg

    def add_ret(self, ret: Argument) -> None:
        """Add a output Argument to Command.

        :param ret: Argument to be added
        :return: Operation success status: True - Success, False - Failure
        """
        self.rets[ret.name] = ret

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """
        args = d["args"]
        rets = d["rets"]

        #self.__dict__.update(make_private(self, d))
        for k,v in self.make_private(self, d).items():
            self.__setattr__(k,v)

        for arg_k, arg_v in args.items():
            arg = Argument()
            arg.decompile(arg_v)
            self.args[arg_k] = arg

        for ret_k, ret_v in rets.items():
            ret = Argument()
            ret.decompile(ret_v)
            self.rets[ret_k] = ret

    def encode(self, src: int, dst: int, args) -> CANMessage:
        common = self.parent.parent.get_common()
        send_cmd = common.get_msg("send_cmd")
        msg = send_cmd.encode(
            {"id": self.id, "dst": dst, "arg1": args[0], "arg2": args[1], "arg3": args[2]},
            src=src
        )
        return msg

    def is_cmd(self, msg: CANMessage) -> bool:
        common = self.parent.parent.get_common()
        send_cmd = common.get_msg("send_cmd")
        return msg.get_msg_id() == send_cmd.id

    def __repr__(self):
        return f"<Cmd name={self.name} id={self.id}>"

    def __hash__(self):
        return hash((self.name, self.id, self.creation_date))



from typing import *
import datetime

from ..can import CANMessage
from .node import Node, Transmit
from .signal import Signal
from ..can import CANMessage
from .utils import normalize

class Message(Transmit):
    """Message node, Represents a CAN message, similar to a DBC message.

    :param name: Name of the Message.
    :param id: FST Message identifier, highest 6 bits of the identifier.
    :param dlc: Message DLC.
    :param signals: Dictionary containing the Message signals.
    :param frequency: Transmission period in millisecond. If 0 message
    isn't automatically sent.
    """

    def __init__(
        self,
        parent: "Device" = None,
        name: str = "",
        id: int = 0,
        dlc: int = 8,
        signals: Dict[str, Signal] = None,
        frequency: int = 0,
        description: str = "",
    ):

        assert parent is not None
        self.parent = parent

        self._id = id or max([msg.id for msg in self.parent.msgs.values()] + [0]) + 1
        self._name = name or ("msg" + str(self.id))
        self._dlc = dlc
        self.signals = {} if signals == None else signals
        self._frequency = frequency
        self._description = description

        self.creation_date = datetime.datetime.now()

    @property
    def name(self) -> str:
        return self._name

    @property
    def id(self) -> int:
        return self._id

    @property
    def dlc(self) -> int:
        return self._dlc

    @property
    def frequency(self) -> int:
        return self._frequency

    @property
    def description(self) -> str:
        return self._description

    @name.setter
    def name(self, name: str) -> None:
        try:
            self._name = name
        except Exception as e:
            return

    @id.setter
    def id(self, id: int) -> None:
        try:
            self._id = int(id)
        except Exception as e:
            return

    @dlc.setter
    def dlc(self, dlc: int) -> None:
        try:
            self._dlc = int(dlc)
        except Exception as e:
            return

    @frequency.setter
    def frequency(self, frequency: int) -> None:
        try:
            self._frequency = int(frequency)
        except Exception as e:
            return

    @description.setter
    def description(self, description: str) -> None:
        try:
            self._description = description
        except Exception as e:
            return


    def compile(self) -> Dict[str, Any]:
        """Transform python class node to its dictionary representation.

        :return: A dictionary containing the node parameters
        """

        sigs = {}
        for k, v in self.signals.items():
            sigs[k] = v.compile()

        d = self.make_public(self, self.filter_private(self.__dict__))
        d["signals"] = sigs
        return d

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """
        signals = d["signals"]
        for k,v in self.make_private(self,d).items():
            self.__setattr__(k,v)

        for key, value in signals.items():
            sig = Signal(self)
            sig.decompile(value)
            self.signals[key] = sig

    def add_signal(self, signal: Signal) -> bool:
        """Add a Signal to Message.

        :param signal: Signal to be added
        :return: Operation success status: True - Success, False - Failure
        """

        if signal == None:
            return False

        if signal.name in self.signals.keys():
            self.signals[signal.name].mux_count += 1
            return True

        self.signals[signal.name] = signal

        return True

    def get_signal(self, name: str) -> Optional[Signal]:
        """Get a Signal from Message by its name.

        :param name: Signal name.
        :return: Signal or None if not found.
        """
        return self.signals.get(name)


    def rm_signal(self, name: str) -> bool:
        """Remove a Signal from Spec.

        :param signal: Signal to be removed.
        """
        if self.get_signal(name) is None:
            return False

        del self.signals[name]
        return True

    def normalize(self):
        """ Update signals dictionary keys."""
        normalize(self.signals)

    def encode(self, signals, src=None):
        assert not ((src is None) and (type(self.parent) == "Common"))
        dev_id = src if src is not None else self.parent.id
        data = 0
        for name, sig in signals.items():
            assert self.signals.get(name) is not None
            data |= self.signals.get(name).encode(sig)

        data16 = [
            (data >> 0) & 0xFFFF,
            (data >> 16) & 0xFFFF,
            (data >> 32) & 0xFFFF,
            (data >> 48) & 0xFFFF]

        sid = (self.id << 5) + self.parent.id
        msg = CANMessage(sid=sid, dlc=self.dlc, data16 = data16, timestamp = 0)

        return msg

    def decode(self, msg: CANMessage):
        mux = ""
        signals = {}
        for name, signal in self.signals.items():
            if signal.mux_count != 1:
                mux = signal.mux
            signals[name] = signal.decode(msg)

        if mux != "":
            mux_value = str(int(signals[mux]))
            signals = {k + (mux_value if k != mux else ""):v for (k,v) in signals.items()}

        return self.name, signals

    def __hash__(self):
        return hash((self.name, self.id, self.creation_date))

    def __repr__(self):
        return f"<Message name={self.name} id={self.id}>"



from typing import *
import datetime

from .node import Node

class Log(Node):
    """Log protocol node.

    :param id: Log integer identifier.
    :param name: Name of the Log node.
    :param n_args: Number of arguments in the Log node.
    :param comment: Description of the Log node
    :param string: Display string for the Log node.
    """

    def __init__(
        self,
        parent: "Spec" = None,
        id: int = 0,
        name: str = "",
        n_args: int = 3,
        comment: str = "",
        string: str = "",
    ):
        self.parent = parent
        assert self.parent is not None

        c = max([log.id for log in self.parent.logs.values()] + [0]) + 1
        self._id = id or c
        self._name = name
        self._n_args = n_args
        self._comment = comment
        self._string = string

        self.creation_date = datetime.datetime.now()

    def compile(self) -> Dict[str, Any]:
        """Transform python class node to its dictionary representation.

        :return: A dictionary containing the node parameters
        """
        return self.make_public(self, self.filter_private(self.__dict__))

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """
        for k, v in self.make_private(self, d).items():
            self.__setattr__(k, v)

    @property
    def id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def n_args(self) -> int:
        return self._n_args

    @property
    def comment(self) -> str:
        return self._comment

    @property
    def string(self) -> str:
        return self._string

    @id.setter
    def id(self, id: int) -> None:
        try:
            self._id = int(id)
        except Exception as e:
            return

    @name.setter
    def name(self, name: str) -> None:
        try:
            self._name = name
        except Exception as e:
            return

    @n_args.setter
    def n_args(self, n_args: int) -> None:
        try:
            self._n_args = int(n_args)
        except Exception as e:
            return

    @comment.setter
    def comment(self, comment: str) -> None:
        try:
            self._comment = comment
        except Exception as e:
            return

    @string.setter
    def string(self, string: str) -> None:
        try:
            self._string = string
        except Exception as e:
            return

    def __hash__(self):
        return hash((self.name, self.id, self.creation_date))

    def __repr__(self):
        return f"<Log name: {self.name}, id: {self.id}, n_args: {self.n_args}>"



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'logwidget.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_LogWidget(object):
    def setupUi(self, LogWidget):
        if not LogWidget.objectName():
            LogWidget.setObjectName(u"LogWidget")
        LogWidget.resize(400, 300)
        LogWidget.setMinimumSize(QSize(300, 0))
        self.verticalLayout = QVBoxLayout(LogWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.groupBox = QGroupBox(LogWidget)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout_3 = QVBoxLayout(self.groupBox)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.scrollArea = QScrollArea(self.groupBox)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.LogScrollContents = QWidget()
        self.LogScrollContents.setObjectName(u"LogScrollContents")
        self.LogScrollContents.setGeometry(QRect(0, 0, 356, 209))
        self.verticalLayout_2 = QVBoxLayout(self.LogScrollContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.logContents = QVBoxLayout()
        self.logContents.setObjectName(u"logContents")

        self.verticalLayout_2.addLayout(self.logContents)

        self.scrollArea.setWidget(self.LogScrollContents)

        self.verticalLayout_3.addWidget(self.scrollArea)

        self.addLogButton = QPushButton(self.groupBox)
        self.addLogButton.setObjectName(u"addLogButton")

        self.verticalLayout_3.addWidget(self.addLogButton)


        self.verticalLayout.addWidget(self.groupBox)


        self.retranslateUi(LogWidget)

        QMetaObject.connectSlotsByName(LogWidget)
    # setupUi

    def retranslateUi(self, LogWidget):
        LogWidget.setWindowTitle(QCoreApplication.translate("LogWidget", u"LogWidget", None))
        self.groupBox.setTitle(QCoreApplication.translate("LogWidget", u"Logs", None))
        self.addLogButton.setText(QCoreApplication.translate("LogWidget", u"Add Log", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'loglayout.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_LogLayout(object):
    def setupUi(self, LogLayout):
        if not LogLayout.objectName():
            LogLayout.setObjectName(u"LogLayout")
        LogLayout.resize(400, 300)
        self.verticalLayout = QVBoxLayout(LogLayout)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.scrollArea = QScrollArea(LogLayout)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.logContents = QWidget()
        self.logContents.setObjectName(u"logContents")
        self.logContents.setGeometry(QRect(0, 0, 380, 280))
        self.verticalLayout_2 = QVBoxLayout(self.logContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.scrollArea.setWidget(self.logContents)

        self.verticalLayout.addWidget(self.scrollArea)


        self.retranslateUi(LogLayout)

        QMetaObject.connectSlotsByName(LogLayout)
    # setupUi

    def retranslateUi(self, LogLayout):
        LogLayout.setWindowTitle(QCoreApplication.translate("LogLayout", u"LogLayout", None))
    # retranslateUi




from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .widgets.messagewidget import Ui_MessageWidget
from .node_details import NodeDetails, FakeParent


class MessageWidget(NodeDetails):
    def load_atts(self, ui, node):
        self.atts = [
            (ui.id, node.id, "id"),
            (ui.name, node.name, "name"),
            (ui.dlc, node.dlc, "dlc"),
            (ui.frequency, node.frequency, "frequency"),
        ]

    def __init__(self, gui, node: "Message", details=None, layout=None):
        NodeDetails.__init__(self, gui, node, details=details, layout=layout)
        self.ui = Ui_MessageWidget()
        self.ui.setupUi(self)

        self.load_atts(self.ui, self.node)
        self.reload()

        self.details_button = self.ui.messageDetailsButton
        self.details_button.clicked.connect(self.open_details)

        self.button_button = self.ui.deleteButton
        self.button_button.clicked.connect(self.delete)



import time
from pathlib import Path

from sqlalchemy import Column, ForeignKey, Integer, Float, String, func, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session
from sqlalchemy import select, update

from .sql import create_session
from .dirs import get_config_dir


ConfigBase = declarative_base()


class File(ConfigBase):
    __tablename__ = "files"

    path = Column(String, primary_key=True)
    time_used = Column(Float)

    def __repr__(self):
        return f"{self.path} {self.time_used}"

    def recent_files(session):
        return session.query(File).order_by(File.time_used.desc()).all()

    def exists(session, path: Path) -> bool:
        files = [x[0] for x in session.query(File.path).all()]
        return str(path) in files

    def access(session, path: Path):
        if not File.exists(session, path):
            file = File(path=str(path), time_used=time.time())
            session.add(file)

        session.execute(
            update(File).where(File.path == str(path)).values(time_used=time.time())
        )

        session.commit()


def config_session():
    config = Path(get_config_dir()) / "config"
    return create_session(config, ConfigBase)


def main():
    session = config_session()

    session.add(File(os.path.abspath("json/fst10e.hjson", time.time())))
    session.commit()


if __name__ == "__main__":
    main()



from .enum import Enum
from .enum_value import EnumValue
from .arg import Argument
from .signal import Signal, SignalValueError
from .message import Message
from .cmd import Command
from .config import Config
from .common import Common
from .device import Device
from .log import Log
from .spec import Spec
from .spec import decompose_id
from .spec import make_sid




from typing import *
import datetime

from .node import Node, Transmit
from ..can import CANMessage

class Config(Transmit):
    """Config node. Represents a Config.

    :param name: Name of the Config.
    :param id: Config identifier.
    :param comment: description of the Config.
    """

    def __init__(
        self,
        parent: "Device" = None,
        name: str = "",
        id: int = 0,
        comment: str = "",
        type: str = "unsigned",
    ):
        self.parent = parent
        self._name = name
        self._id = int(id)
        self._comment = comment
        self._type = type

        self.creation_date = datetime.datetime.now()

    @property
    def name(self) -> str:
        return self._name
 
    @property
    def id(self) -> int:
        return int(self._id)

    @property
    def comment(self) -> str:
        return self._comment

    @property
    def type(self) -> str:
        return self._type

    @name.setter
    def name(self, name: str) -> None:
        try:
            self._name = name
        except Exception as e:
            return

    @id.setter
    def id(self, id: int) -> None:
        try:
            self.id = int(id)
        except Exception as e:
            return

    @comment.setter
    def comment(self, comment: str) -> None:
        try:
            self.comment = comment
        except Exception as e:
            return

    @type.setter
    def type(self, type: str) -> None:
        self.type = type

    def compile(self) -> Dict[str, Any]:
        """Transform python class node to its dictionary representation.

        :return: A dictionary containing the node parameters
        """

        return self.make_public(self, self.filter_private(self.__dict__))

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """
        #self.__dict__.update(make_private(self, d))
        for k,v in self.make_private(self, d).items():
            self.__setattr__(k,v)

    def normalize(self):
        return

    def encode_set(self, src: int, dst: int, value: int) -> CANMessage:
        common = self.parent.parent.get_common()
        req_set = common.get_msg("req_set")
        msg = req_set.encode({
            "id": self.id,
            "dst": dst,
            "data": value
        })

        return msg

    def encode_get(self, src: int, dst: int) -> CANMessage:
        common = self.parent.parent.get_common()
        req_set = common.get_msg("req_get")
        msg = req_set.encode({
            "id": self.id,
            "dst": dst,
        })

        return msg

    def __hash__(self):
        return hash((self.name, self.id, self.creation_date))



from typing import *
import datetime

from .node import Node

class Argument(Node):
    """Argument node. Represents a Command Argument.

    :param name: Name of the Argument.
    :param id: Argument identifier.
    :param comment: description of the Argument.
    """

    def __init__(
        self,
        parent: "Command" = None,
        name: str = "",
        id: int = 0,
        comment: str = "",
        type: str = "unsigned",
    ):
        self.parent = parent
        self._name = name
        self._id = id
        self._comment = comment
        self._type = type

        self.creation_date = datetime.datetime.now()

    @property
    def name(self) -> str:
        return self._name

    @property
    def comment(self) -> str:
        return self._comment

    @property
    def id(self) -> int:
        return self._id

    @property
    def type(self) -> str:
        return self._type

    @name.setter
    def name(self, name: str) -> None:
        try:
            self._name = name
        except Exception as e:
            return

    @comment.setter
    def comment(self, comment: str) -> None:
        try:
            self._comment = comment
        except Exception as e:
            return

    @id.setter
    def id(self, id: int) -> None:
        try:
            self._id = int(id)
        except Exception as e:
            return

    @type.setter
    def type(self, type: str) -> None:
        self._type = type

    def compile(self) -> Dict[str, Any]:
        """Transform python class node to its dictionary representation.

        :return: A dictionary containing the node parameters
        """
        d = self.make_public(self, self.filter_private(self.__dict__))
        return d

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """
        #self.__dict__.update(make_private(self, d))
        for k,v in self.make_private(self,d).items():
            self.__setattr__(k,v)



from .node_details import NodeDetails, FakeParent
from .widgets.signaldetails import Ui_SignalDetails


class SignalDetails(NodeDetails):
    def load_atts(self, ui, node):
        self.atts = [
            (ui.nameEdit, node.name, "name"),
            (ui.startEdit, node.start, "start"),
            (ui.lengthEdit, node.length, "length"),
            (ui.muxEdit, node.mux, "mux"),
            (ui.muxCountEdit, node.mux_count, "mux_count"),
            (ui.typeEdit, node.type, "type"),
            (ui.commentEdit, node.comment, "comment"),
            (ui.minValueEdit, node.min_value, "min_value"),
            (ui.maxValueEdit, node.max_value, "max_value"),
            (ui.byteOrderEdit, node.byte_order, "byte_order"),
            (ui.scaleEdit, node.scale, "scale"),
            (ui.offsetEdit, node.offset, "offset"),
            (ui.aliasEdit, node.alias, "alias"),
        ]

    def __init__(self, gui, node: "Signal", parent):
        NodeDetails.__init__(self, gui, node, parent)

        ui = self.ui = Ui_SignalDetails()
        self.ui.setupUi(self)

        self.load_atts(ui, node)
        self.connect_atts()

        self.ui.signalDeleteButton.clicked.connect(self.delete)

        self.reload()
        self.setVisible(False)

        self.children = []



import pytest

from ...specs import *
from ...can import CANMessage

@pytest.fixture(scope="session")
def device():
    spec = Spec()
    dev = Device(parent = spec)
    return dev

@pytest.fixture(scope="session")
def message(device):
    msg = Message(parent = device, id=1)
    msg.add_signal(Signal(parent=msg, name="test1", start=0, length=8))
    msg.add_signal(Signal(parent=msg, name="test5", start=8, length=8))
    msg.add_signal(Signal(parent=msg, name="test2", start=16, length=16))
    msg.add_signal(Signal(parent=msg, name="test3", start=32, length=16))
    msg.add_signal(Signal(parent=msg, name="test4", start=48, length=16))
    return msg

#@pytest.fixture(scope="session")

def test_encode(message):
    msg = message.encode({"test1": 10, "test5": 10})
    assert type(msg) is CANMessage
    print(msg)
    assert msg.get_data16()[0] == 2570

#def test_decode(message, can_msg):
#    



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'enumwidget.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_EnumWidget(object):
    def setupUi(self, EnumWidget):
        if not EnumWidget.objectName():
            EnumWidget.setObjectName(u"EnumWidget")
        EnumWidget.resize(400, 300)
        self.verticalLayout = QVBoxLayout(EnumWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.groupBox = QGroupBox(EnumWidget)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout_3 = QVBoxLayout(self.groupBox)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.scrollArea = QScrollArea(self.groupBox)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.LogScrollContents = QWidget()
        self.LogScrollContents.setObjectName(u"LogScrollContents")
        self.LogScrollContents.setGeometry(QRect(0, 0, 368, 208))
        self.verticalLayout_2 = QVBoxLayout(self.LogScrollContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.enumContents = QVBoxLayout()
        self.enumContents.setObjectName(u"enumContents")

        self.verticalLayout_2.addLayout(self.enumContents)

        self.scrollArea.setWidget(self.LogScrollContents)

        self.verticalLayout_3.addWidget(self.scrollArea)

        self.addEnumButton = QPushButton(self.groupBox)
        self.addEnumButton.setObjectName(u"addEnumButton")

        self.verticalLayout_3.addWidget(self.addEnumButton)


        self.verticalLayout.addWidget(self.groupBox)


        self.retranslateUi(EnumWidget)

        QMetaObject.connectSlotsByName(EnumWidget)
    # setupUi

    def retranslateUi(self, EnumWidget):
        EnumWidget.setWindowTitle(QCoreApplication.translate("EnumWidget", u"EnumWidget", None))
        self.groupBox.setTitle(QCoreApplication.translate("EnumWidget", u"Enums", None))
        self.addEnumButton.setText(QCoreApplication.translate("EnumWidget", u"Add Enums", None))
    # retranslateUi




from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .widgets.logwidget import Ui_LogWidget
from .node_details import NodeDetails, FakeParent
from .log_details import LogDetails

from .undo_redo import UndoAdd

from ..specs import Log


class LogWidget(QWidget):
    def __init__(self, gui, parent):
        self.Window = True
        QWidget.__init__(self)
        self.ui = Ui_LogWidget()
        self.ui.setupUi(self)

        self.parent = parent
        self.gui = gui

        self.atts = []

        self.children = []

        self.reload()

        self.details = None

        self.ui.addLogButton.clicked.connect(self.add_log)
        logs = [log for log in self.parent.logs.values()]
        for log in logs:
            self.add_log(log)

    def reload(self):
        return

    def save(self):
        for child in self.children:
            child.save()

        self.gui.reload()

    def add_log(self, log=None):
        new_log = log is None or type(log) is bool
        if log == None or log == False:
            log = Log(self.parent)

        self.parent.logs[log.name] = log
        log_widget = LogDetails(self.gui, log, FakeParent())
        self.ui.logContents.addWidget(log_widget)
        self.children.append(log_widget)
        
        if new_log:
            undo_action = UndoAdd(log, log_widget, log_widget.delete, self.add_log)
            self.gui.undo_redo.push(undo_action)



import json
from typing import *
from math import *
import queue
from threading import Thread
from result import Ok, Err
import atexit

from fcp.specs import *
from fcp.can import CANMessage


def encode_signal(signal, value):
    value = (value - signal.offset) / signal.scale
    return (int(value) & bitmask(signal.length)) << signal.start


def conv_endianess(value: int, signal: Signal):
    length = signal.length / 8
    log2_length = log2(length)

    if length == 1:
        return value

    if log2_length != int(log2_length):
        return value

    if signal.byte_order != "big_endian":
        return value

    length = int(length)

    def swap64(x):
        return int.from_bytes(
            value.to_bytes(8, byteorder="little"), byteorder="big", signed=False
        )

    def swap32(x):
        return int.from_bytes(
            value.to_bytes(4, byteorder="little"), byteorder="big", signed=False
        )

    def swap16(x):
        return int.from_bytes(
            value.to_bytes(2, byteorder="little"), byteorder="big", signed=False
        )

    if length == 8:
        return swap64(value)
    if length == 4:
        return swap32(value)
    if length == 2:
        return swap16(value)

class Fcp:
    def __init__(self, spec):
        self.messages = {}

        if type(spec) is str:
            with open(spec) as f:
                j = json.loads(f.read())
            spec = Spec()
            spec.decompile(j)
            self.spec = spec
        elif type(spec) is Spec:
            self.spec = spec


        for device in spec.devices.values():
            for msg in device.msgs.values():
                self.messages[msg.name] = msg

        for msg in spec.common.msgs.values():
            self.messages[msg.name] = msg

    def encode_msg(
        self, msg_name: str, signals: Dict[str, float], src: int = None) -> CANMessage:
        msg = self.messages.get(msg_name)
        return msg.encode(signals, src)

    def decode_msg(self, msg: CANMessage):
        fcp_msg = self.find_msg(msg)
        if fcp_msg is None:
            return "", {}

        return fcp_msg.decode(msg)

    def encode_cmd(self, src: int, dst: str, name: str, args) -> CANMessage:
        device = self.find_device(dst)
        if device is None:
            return Err(f"{dst} device not found")

        cmd = device.get_cmd(name)
        if cmd is None:
            return Err(f"{dst}/{name} cmd not found")

        return Ok(cmd.encode(src, device.id, args))

    def find_msg(self, msg):
        dev_id, msg_id = decompose_id(msg.sid)

        for dev in self.spec.devices.values():
            if dev.id == dev_id:
                for msg in dev.msgs.values():
                    if msg.id == msg_id:
                        return msg

        for msg in self.spec.common.msgs.values():
            if msg.id == msg_id:
                return msg

    def find_device(self, id):
        for dev in self.spec.devices.values():
            if isinstance(id, int) and dev.id == id:
                return dev
            if isinstance(id, str) and dev.name == id:
                return dev

        return None

    def find_config(self, dev, id):
        for name, cfg in dev.cfgs.items():
            if isinstance(id, str) and name == id:
                return cfg

        return None


    def decode_log(self, signal):
        for name, log in self.spec.logs.items():
            if log.id == signal["id"]:
                return log.string

        return ""

    def decode_get(self, signal):
        dev = self.find_device(signal["dst"])
        if dev == None:
            return None

        for name, config in dev.cfgs.items():
            if config.id == signal["id"]:
                return {config.name: signal["data"]}

    def encode_get(self, src: int, dst: int, config: int):
        device = self.find_device(dst)
        if device is None:
            return Err(f"{dst} device not found")

        cfg = device.get_cfg(config)
        if cfg is None:
            return Err(f"{dst}/{name} cfg not found")

        return Ok(cfg.encode_get(src, device.id))

    def decode_set(self, signal):
        dev = self.find_device(signal["dst"])
        if dev == None:
            return None

        for name, config in dev.cfgs.items():
            if config.id == signal["id"]:
                return {"device": device.name, config.name: signal["data"]}

    def encode_set(self, src: int, dst: str, config: str, value: int):
        device = self.find_device(dst)
        if device is None:
            return Err(f"{dst} device not found")

        cfg = device.get_cfg(config)
        if cfg is None:
            return Err(f"{dst}/{name} cfg not found")

        return Ok(cfg.encode_set(src, device.id, value))

class Proxy():
    def __init__(self, socket, addrs):
        self.socket = socket
        self.addrs = []

        self.cmds = {}

    def recv(self) -> CANMessage:
        msg, addr = self.socket.recvfrom(1024)
        return Ok(CANMessage.decode_json(msg))

    def send(self, msg: CANMessage):
        for addr in self.addrs:
            self.socket.sendto(msg.encode_json(), addr)

class FcpCom():
    def __init__(self, fcp: Fcp, proxy: Proxy, id: int = 31):
        self.fcp = fcp
        self.proxy = proxy
        self.id = id

        self.terminate = False
        self.cmds = {}
        self.sets = {}
        self.gets = {}
        self.q = queue.Queue()

    def start(self):
        self.thread = Thread(target=self.run, daemon=True)
        self.thread.start()

    def stop(self):
        self.terminate = True
        self.thread.join()

    def cmd(self, dst, name, args):
        if self.cmds.get((dst, name)) is None:
            self.cmds[(dst, name)] = queue.Queue()

        msg = self.fcp.encode_cmd(self.id, dst, name, args)
        if msg.is_err():
            return msg
        msg = msg.ok()
        self.proxy.send(msg)

        try:
            r = self.cmds[(dst, name)].get(timeout=2)
            return Ok(r)
        except queue.Empty as e:
            return Err("Timeout")

    def set(self, dst, name, value):
        if self.sets.get((dst, name)) is None:
            self.sets[(dst, name)] = queue.Queue()

        msg = self.fcp.encode_set(self.id, dst, name, value)
        if msg.is_err():
            return msg
        msg = msg.ok()
        self.proxy.send(msg)

        try:
            self.sets[(dst, name)].get(timeout=2)
            return Ok()
        except queue.Empty as e:
            return Err("Timeout")

    def get(self, dst, name):
        if self.gets.get((dst, name)) is None:
            self.gets[(dst, name)] = queue.Queue()

        msg = self.fcp.encode_get(self.id, dst, name)
        if msg.is_err():
            return msg
        msg = msg.ok()
        self.proxy.send(msg)

        try:
            r = self.gets[(dst, name)].get(timeout=2)
            return Ok(r)
        except queue.Empty as e:
            return Err("Timeout")


    def run(self):
        while True:
            if self.terminate == True:
                break
            r = self.proxy.recv()
            if r.is_err():
                continue
            msg = r.ok()
            name, signals = self.fcp.decode_msg(msg)

            self.q.put((name, signals))

            if name == "return_cmd":
                dev = self.fcp.spec.get_device(msg.get_dev_id())
                cmd = dev.get_cmd(signals["id"])
                rets = (signals["ret1"], signals["ret2"], signals["ret3"])
                q = self.cmds.get((dev.name, cmd.name))
                if q is not None:
                    q.put(rets)
            elif name == "ans_set":
                dev = self.fcp.spec.get_device(msg.get_dev_id())
                cfg = dev.get_cfg(signals["id"])
                q = self.sets.get((dev.name, cfg.name))
                if q is not None:
                    q.put(None)
            elif name == "ans_get":
                dev = self.fcp.spec.get_device(msg.get_dev_id())
                cfg = dev.get_cfg(signals["id"])
                ret = signals["data"]
                q = self.gets.get((dev.name, cfg.name))
                if q is not None:
                    q.put(ret)
                pass




from typing import *

def normalize(xs: Dict[str, Any], key: Callable[[Any], str] = None):
    """Update xs dictionary keys according to key.
        By default key is `lambda x : x.name`

    :param xs: Dictionary containing spec node
    :param key: Function that returns the key for a particular node
    """

    if key == None:
        key = lambda x: x.name

    aux = []

    for k, x in xs.items():
        if k != key(x):
            aux.append((k, key(x)))

    for k, key in aux:
        xs[key] = xs[k]
        del xs[k]



from typing import *
import datetime

from .node import Node

class EnumValue(Node):
    """Fcp EnumValue. C lookalike for FCP type definitions with name-value
    associations.
    """

    def __init__(self, parent: "Enum" = None) -> None:
        self.parent = parent

        assert self.parent is not None

        c = max([value.value for value in self.parent.enumeration.values()] + [0]) + 1
        self._name = ""
        self._value = c

        self.creation_date = datetime.datetime.now()

    def compile(self) -> Dict[str, Any]:
        d = make_public(self, filter_private(self.__dict__))
        return d

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """

        for k,v in self.make_private(self, d).items():
            self.__setattr__(k,v)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, value: int) -> None:
        self._value = int(value)

    def __hash__(self):
        return hash((self._name, self.creation_date))

    def __repr__(self):
        return "name: {}".format(self.name)



from typing import *
import copy
import datetime
import time

from . import *
from .utils import normalize


def handle_key_not_found(d: dict, key: str):
    return d.get(key).items() if d.get(key) != None else []

class Spec:
    """FCP root node. Holds all Devices, Messages, Signals, Logs, Configs,
    Commands and Arguments.
    """

    def __init__(self) -> None:
        self.devices = {}
        self.common = Common()
        self.logs = {}
        self.enums = {}
        self.version = "0.2"
        self.parent = None
        self.name = ""

    def add_device(self, device: "Device") -> bool:
        """Add a Device to Spec.

        :param device: Device to be added
        :return: Operation success status: True - Success, False - Failure
        """
        if device is None:
            return False

        if device.name in self.devices.keys():
            return False

        self.devices[device.name] = device

        return True

    def add_enum(self, enum: "Enum") -> bool:
        """Add a Enum to Spec.

        :param enum: Enum to be added
        :return: Operation success status: True - Success, False - Failure
        """

        if enum == None:
            return False

        if enum.name in self.enums.keys():
            return False

        self.enums[enum.name] = enum

        return True

    def add_log(self, log: "Log") -> bool:
        """Add a Log to Spec.

        :param log: Log to be added
        :return: Operation success status: True - Success, False - Failure
        """

        if log == None:
            return False

        if log.name in self.logs.keys():
            return False

        self.logs[log.name] = log

        return True

    def get_common(self) -> Optional["Common"]:
        return self.common

    def get_device(self, name: Union[str, int]) -> Optional["Device"]:
        """Get a Device from Spec by its name.

        :param name: Device name.
        :return: Device or None if not found.
        """

        if type(name) == str:
            return self.devices.get(name)
        if type(name) == int:
            for device in self.devices.values():
                if device.id == name:
                    return device

    def get_log(self, name: str) -> Optional["Log"]:
        """Get a Log from Spec by its name.

        :param name: Log name.
        :return: Log or None if not found.
        """
        return self.logs.get(name)

    def get_signals(self) -> List[Signal]:
        for dev in self.devices.values():
            for msg in dev.msgs.values():
                for signal in msg.signals.values():
                    yield signal

    def rm_node(self, node: Any) -> None:
        """Remove a node from Spec.

        :param node: node to be removed.
        """
        if type(node) == Device:
            self.rm_device(node)
        elif type(node) == Message:
            self.rm_message(node)
        elif type(node) == Signal:
            self.rm_signal(node)
        elif type(node) == Config:
            self.rm_config(node)
        elif type(node) == Command:
            self.rm_cmd(node)
        elif type(node) == Log:
            self.rm_log(node)
        elif type(node) == Enum:
            self.rm_enum(node)
        elif type(node) == EnumValue:
            self.rm_enum_value(node)

    def rm_device(self, device: "Device") -> None:
        """Remove a Device from Spec.

        :param device: Device to be removed.
        """
        devs = [dev.name for dev in self.devices.values() if dev == device]

        for name in devs:
            del self.devices[name]

    def rm_message(self, message: "Message") -> None:
        """Remove a Message from Spec.

        :param message: Message to be removed.
        """
        for dev in self.devices.values():
            msgs = []
            for msg in dev.msgs.values():
                if msg == message:
                    msgs.append(msg.name)

            for name in msgs:
                dev.rm_msg(name)

    def rm_signal(self, signal: "Signal") -> None:
        """Remove a Signal from Spec.

        :param signal: Signal to be removed.
        """

        for dev in self.devices.values():
            for msg in dev.msgs.values():
                signals = []
                for sig in msg.signals.values():
                    if sig is signal:
                        signals.append(signal.name)

                for name in signals:
                    msg.rm_signal(name)

    def rm_config(self, config: "Config") -> None:
        """Remove a Config from Spec.

        :param config: Config to be removed.
        """

        for dev in self.devices.values():
            cfgs = []
            for cfg in dev.cfgs.values():
                if cfg is config:
                    cfgs.append(cfg.name)

            for name in cfgs:
                dev.rm_cfg(name)

    def rm_cmd(self, command: "Command") -> None:
        """Remove a Command from Spec.

        :param command: Command to be removed.
        """

        for dev in self.devices.values():
            cmds = []
            for cmd in dev.cfgs.values():
                if cmd is command:
                    cmds.append(cmd)

            for name in cmds:
                dev.rm_cmd(name)

    def rm_log(self, log):
        """Remove a Log from Spec.

        :param log: Log to be removed.
        """
        logs = []

        for l in self.logs.values():
            if log is l:
                logs.append(l.name)

        for name in logs:
            del self.logs[name]

    def rm_enum(self, enum):
        """Remove a Enum from Spec.

        :param log: Enum to be removed.
        """
        enums = []

        for l in self.enums.values():
            if enum is l:
                enums.append(l.name)

        for name in enums:
            del self.enums[name]

    def rm_enum_value(self, enum_value):
        """Remove a Enum from Spec.

        :param log: EnumValue to be removed.
        """

        for e in self.enums.values():
            enum_values = []
            for ev in e.enumeration.values():
                if enum_value is ev:
                    enum_values.append(ev.name)

            for name in enum_values:
                del e.enumeration[name]

    def compile(self) -> Dict[str, Any]:
        """Transform python class node to its dictionary representation.

        :return: A dictionary containing the node parameters
        """


        d = {"devices": {}, "logs": {}, "enums": {}}

        for dev_k, dev_v in self.devices.items():
            d["devices"][dev_k] = dev_v.compile()


        for log in self.logs.values():
            d["logs"][log.name] = log.compile()



        for enum_k, enum_v in self.enums.items():
            d["enums"][enum_k] = enum_v.compile()


        d["common"] = self.common.compile()
        d["version"] = self.version


        return d

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """
        d = copy.copy(d)
        self.devices = {}
        self.logs = {}
        self.common.decompile(d["common"])

        for k, v in handle_key_not_found(d, "devices"):
            dev = Device(self)
            dev.decompile(v)
            self.devices[k] = dev

        for k, v in handle_key_not_found(d, "logs"):
            log = Log(self)
            log.decompile(v)
            self.logs[k] = log

        for k, v in handle_key_not_found(d, "enums"):
            enum = Enum(self)
            enum.decompile(v)
            self.enums[k] = enum

        self.version = d["version"]

    def normalize(self):
        """ Update devices and logs dictionary keys.  """

        normalize(self.devices)
        normalize(self.logs)

        normalize(self.enums)
        for enum in self.enums.values():
            enum.normalize()

        for key, dev in self.devices.items():
            dev.normalize()

    def __repr__(self):

        msg_count = 0
        cfg_count = 0
        cmd_count = 0
        sig_count = 0

        for dev in self.devices.values():
            for msg in dev.msgs.values():
                msg_count += 1
                for sig in msg.signals.values():
                    sig_count += 1
            for cfg in dev.cfgs.values():
                cfg_count += 1
            for cmd in dev.cmds.values():
                cmd_count += 1

        return f"(Spec: devs={len(self.devices)}, msgs={msg_count}, sigs={sig_count}, logs={len(self.logs)}, cfgs={cfg_count}, cmds={cmd_count})"


def decompose_id(sid: int) -> Tuple[int, int]:
    """ Find the dev_id and the msg_id from the sid."""
    return sid & 0x1F, (sid >> 5) & 0x3F


def make_sid(dev_id: int, msg_id: int) -> int:
    """ Find the sid from the dev_id and the msg_id """
    return (msg_id << 5) + dev_id



# Documentation generator
# Generate HTML from json
import os
from jinja2 import Template

from .docs_css import css

signals_template = Template("""
<html>
<head>
<style>
#searchbar {
  background-image: url('css/searchicon.png'); /* Add a search icon to input */
  background-position: 10px 12px; /* Position the search icon */
  background-repeat: no-repeat; /* Do not repeat the icon image */
  width: 100%; /* Full-width */
  font-size: 14px; /* Increase font-size */
  padding: 12px 20px 12px 40px; /* Add some padding */
  border: 1px solid #ddd; /* Add a grey border */
  margin-bottom: 12px; /* Add some space below the input */
}

#signals {
  border-collapse: collapse; /* Collapse borders */
  width: 100%; /* Full-width */
  border: 1px solid #ddd; /* Add a grey border */
  font-size: 14px; /* Increase font-size */
}

#signals th, #signals td {
  text-align: left; /* Left-align text */
  padding: 12px; /* Add padding */
}

#signals tr {
  /* Add a bottom border to all table rows */
  border-bottom: 1px solid #ddd;
}

#signals tr.header, #signals tr:hover {
  /* Add a grey background color to the table header and on hover */
  background-color: #ffdd11;
}
</style>
<script>
function search_function() {
  // Declare variables
  var input, filter, table, tr, td, i, txtValue;
  input = document.getElementById("searchbar");
  filter = input.value.toUpperCase();
  table = document.getElementById("signals");
  tr = table.getElementsByTagName("tr");

  // Loop through all table rows, and hide those who don't match the search query
  for (i = 0; i < tr.length; i++) {
    td0 = tr[i].getElementsByTagName("td")[0];
    td1 = tr[i].getElementsByTagName("td")[1];
    if (td0 || td1) {
      txtValue = td0.textContent || td0.innerText;
      txtValue += td1.textContent || td1.innerText;
      if (txtValue.toUpperCase().indexOf(filter) > -1) {
        tr[i].style.display = "";
      } else {
        tr[i].style.display = "none";
      }
    }
  }
}
</script>
</head>
<body>
<br>
<input type="text" id="searchbar" onkeyup="search_function()" placeholder="Search">
<title>Signals</title>
<ul>
<table id="signals">
<tr class="header">
<th style="width:10%;">Name</th>
<th style="width:70%;">Description</th>
</tr>
{% for sig in signals %}
<tr>
<td>{{sig.name}}</td>
<td>{{sig.comment}}</td>
</tr>
{% endfor %}
</table>
</ul>
</body>
</html>
""")

def markdown(spec, root):
    main = "% FCP Docs\n"
    # markdown += "# FCP Docs\n"
    main += "## Signal list\n"
    main += "[Signals list](signals.html)\n\n"

    main += f"## Logs\n"
    for log in sorted(spec.logs.values(), key=lambda x: x.id):
        main += f"* {log.id}: [{log.name}]({root}/log.md#{log.name})\n"

    main += "\n"

    main += f"## Devices\n"
    for dev in sorted(spec.devices.values(), key=lambda x: x.id):
        main += f"* [{dev.id}: {dev.name}]({root}/{dev.name.lower()}.md)\n"
        for msg in sorted(dev.msgs.values(), key=lambda x: x.id):
            main += f"  + [{msg.id}: {msg.name}]({root}/{dev.name.lower()}.md#{msg.name.lower()})\n"
    main += "\n"

    main += f"## Configs\n"
    for dev in sorted(spec.devices.values(), key=lambda x: x.id):
        main += f"* [{dev.id}: {dev.name}]({root}/{dev.name.lower()}.md)\n"
        for cfg in sorted(dev.cfgs.values(), key=lambda x: x.id):
            main += f"  + [{cfg.id}: {cfg.name}]({root}/{dev.name.lower()}.md#{cfg.name.lower()})\n"
    main += "\n"

    main += f"## Commands\n"
    for dev in sorted(spec.devices.values(), key=lambda x: x.id):
        main += f"* [{dev.id}: {dev.name}]({root}/{dev.name.lower()}.md)\n"
        for cmd in sorted(dev.cmds.values(), key=lambda x: x.id):
            main += f"  + [{cmd.id}: {cmd.name}]({root}/{dev.name.lower()}.md#{cmd.name.lower()})\n"
    main += "\n"

    devices = []
    for dev in spec.devices.values():
        device = f"% {dev.name} \n"
        device += f"[index]({root}/index.md)\n\n"
        device += "## Messages\n"
        for msg in sorted(dev.msgs.values(), key=lambda x: x.id):
            device += f"### {msg.id}. {msg.name}\n"
            device += f"* dlc: {msg.dlc}\n"
            device += f"* frequency: {msg.frequency}\n\n"
            for sig in sorted(msg.signals.values(), key=lambda x: x.start):
                device += f"#### {sig.name}\n"
                if sig.comment != "":
                    device += f"_{sig.comment}_\n\n"
                device += f"* start: {sig.start}\n"
                device += f"* length: {sig.length}\n"

                if sig.byte_order != "little_endian":
                    device += f"* byte_order: {sig.byte_order}\n"

                if not (sig.scale == 1 and sig.offset == 0):
                    device += f"* scale: {sig.scale}\n"
                    device += f"* offset: {sig.offset}\n"

                device += f"* type: {sig.type}\n"
                if sig.unit != "":
                    device += f"* unit: {sig.unit}\n\n"

                if not (sig.min_value == 0 and sig.max_value == 0):
                    device += f"* min_value: {sig.min_value}\n"
                    device += f"* max_value: {sig.max_value}\n"

                if not (sig.mux_count == 1 and sig.mux == ""):
                    device += f"* mux: {sig.mux}\n"
                    device += f"* mux_count: {sig.mux_count}\n"
                device += "\n"

        device += f"## Configs \n"
        for cfg in sorted(dev.cfgs.values(), key=lambda x: x.id):
            device += f"### {cfg.name} ({cfg.id}) : {cfg.type}\n"
            if not cfg.comment == "":
                device += f"_{cfg.comment}_\n"
            device += "\n"

        device += f"## Commands \n"
        for cmd in sorted(dev.cmds.values(), key=lambda x: x.id):
            device += f"### {cmd.id}. {cmd.name}\n"
            if not cmd.comment == "":
                device += f"_{cmd.comment}_\n"

            if len(cmd.args) > 0:
                device += "\n#### Arguments\n"
                for arg in cmd.args.values():
                    device += f" * {arg.name} ({arg.id}): {arg.type}\n"

            if len(cmd.rets) > 0:
                device += "\n#### Returns\n"
                for ret in cmd.rets.values():
                    device += f" * {ret.name} ({arg.id}): {arg.type}\n"

            device += "\n"

        devices.append(device)

    log_md = "% Logs\n"
    log_md += f"[index]({root}/index.md)\n\n"
    for log in sorted(spec.logs.values(), key=lambda x: x.id):
        log_md += f"## {log.name}\n"
        if log.comment != "":
            log_md += f"_{log.comment}_\n\n"

        log_md += f"* {log.string}\n"
        log_md += f"* {log.id}\n"
        log_md += f"* {log.n_args}\n"
        log_md += "\n"

    signals = spec.get_signals()
    signals = [sig.compile() for sig in signals]
    signals_txt = signals_template.render({"signals": signals})
    return main, log_md, devices, signals_txt


def check_out_dir(out):
    """Create output directory.

    :out: output directory path.
    :return: Failure and error message.
    """

    if not os.path.isdir(out):
        try:
            os.makedirs(out)
        except Exception as e:
            return False, str(e)

    return True, ""


def generate_docs(spec, out, link_location, logger):

    r, msg = check_out_dir(out)
    if not r:
        logger.error(msg)
        return

    root = link_location
    logger.info("Generate docs")

    build_sh = """
    for file in *.md; do
            pandoc -s --css=pandoc.css -f markdown --to=html5 "$file" -o "$(basename "$file" .md).html" --lua-filter=links-to-html.lua; 
    done
    """

    links_to_html = """
-- links-to-html.lua
function Link(el)
  el.target = string.gsub(el.target, "%.md", ".html")
  return el
end
"""

    main, log_md, devices, signals = markdown(spec, link_location)

    with open(os.path.join(out, "signals.html"), "w") as f:
        f.write(signals)

    with open(os.path.join(out, "index.md"), "w") as _f:
        _f.write(main)

    for device, name in zip(devices, spec.devices.keys()):
        with open(os.path.join(out, name + ".md"), "w") as _f:
            _f.write(device)

    with open(os.path.join(out, "log.md"), "w") as _f:
        _f.write(log_md)

    with open(os.path.join(out, "build.sh"), "w") as _f:
        _f.write(build_sh)

    os.chmod(os.path.join(out, "build.sh"), 0o755)

    with open(os.path.join(out, "pandoc.css"), "w") as _f:
        _f.write(css)

    with open(os.path.join(out, "links-to-html.lua"), "w") as _f:
        _f.write(links_to_html)

    os.system(f"cd {out}; ./build.sh")

    logger.info("Done")



from typing import *
import datetime

from .node import Node
from .message import Message
from .cmd import Command
from .config import Config
from .utils import normalize

class Device(Node):
    """Device node, Represents a CAN device.

    :param name: Name of the Device.
    :param id: FST Device identifier, lowest 5 bits of the identifier.
    :param msgs: Dictionary containing the Device messages.
    :param cmds: Dictionary containing the Device commands.
    :param cfgs: Dictionary containing the Device configs.
    isn't automatically sent.
    """

    def __init__(
        self,
        parent: "Spec" = None,
        name: str = "",
        id: int = 0,
        msgs: Dict[str, Message] = None,
        cmds: Dict[str, Command] = None,
        cfgs: Dict[str, Config] = None,
    ):

        self.parent = parent
        assert self.parent is not None
        c = max([dev.id for dev in self.parent.devices.values()] + [0]) + 1

        self._name = name or ("device" + str(c))
        self._id = id or c
        self.msgs = {} if msgs == None else msgs
        self.cmds = {} if cmds == None else cmds
        self.cfgs = {} if cfgs == None else cfgs

        self.creation_date = datetime.datetime.now()

    @property
    def name(self) -> str:
        return self._name

    @property
    def id(self) -> int:
        return self._id

    @name.setter
    def name(self, name: str) -> None:
        try:
            self._name = name
        except Exception as e:
            return

    @id.setter
    def id(self, id: int) -> None:
        print("set id")
        try:
            self._id = int(id)
        except Exception as e:
            return


    def add_cmd(self, cmd: Command) -> None:
        self.cmds[cmd.name] = cmd

    def add_cfg(self, cfg: Config) -> None:
        self.cfgs[cfg.name] = cfg

    def compile(self) -> Dict[str, Any]:
        """Transform python class node to its dictionary representation.

        :return: A dictionary containing the node parameters
        """

        msgs = {}
        cmds = {}
        cfgs = {}

        for msg_k, msg_v in self.msgs.items():
            msgs[msg_k] = msg_v.compile()

        for cmd_k, cmd_v in self.cmds.items():
            cmds[cmd_k] = cmd_v.compile()

        for cfg_k, cfg_v in self.cfgs.items():
            cfgs[cfg_k] = cfg_v.compile()

        att = self.make_public(self, self.filter_private(self.__dict__))
        att["msgs"] = msgs
        att["cmds"] = cmds
        att["cfgs"] = cfgs

        return att

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """

        msgs = d["msgs"]
        cmds = d["cmds"]
        cfgs = d["cfgs"]

        for k, v in self.make_private(self, d).items():
            f = self.__setattr__(k, v)

        #self.__dict__.update()

        for k, v in msgs.items():
            msg = Message(self)
            msg.decompile(v)
            self.msgs[k] = msg

        for k, v in cmds.items():
            cmd = Command(self)
            cmd.decompile(v)
            self.cmds[k] = cmd

        for k, v in cfgs.items():
            cfg = Config(self)
            cfg.decompile(v)
            self.cfgs[k] = cfg

    def add_msg(self, msg: Message) -> bool:
        """Add a Message to Device.

        :param msg: Message to be added.
        :return: Operation success status: True - Success, False - Failure
        """
        if msg == None:
            return False

        if msg.name in self.msgs.keys():
            return False

        self.msgs[msg.name] = msg

        return True

    def get_msg(self, name: str) -> Optional[Message]:
        """Get a Message from Device by its name.

        :param name: Message name.
        :return: Message or None if not found.
        """
        return self.msgs.get(name)

    def rm_msg(self, name: str) -> bool:
        """Remove a Message from Device.

        :param name: Name of the Message to be removed.
        """
        if self.get_msg(name) is None:
            return False

        del self.msgs[name]
        return True

    def rm_cmd(self, name: str) -> bool:
        """Remove a Command from Device.

        :param name: Name of the Command to be removed.
        """
        if self.get_cmd(name) is None:
            return False

        del self.cmds[name]
        return True

    def get_cfg(self, name: str) -> Optional[Config]:
        """Get a Config from Device by its name.

        :param name: Config name.
        :return: Config or None if not found.
        """

        if type(name) is str:
            return self.cfgs.get(name)
        if type(name) is int or type(name) is float:
            for cfg in self.cfgs.values():
                if cfg.id == int(name):
                    return cfg

    def rm_cfg(self, name: str) -> bool:
        """Remove a Config from Device.

        :param name: Name of the Config to be removed.
        """
        if self.get_cfg(name) is None:
            return False

        del self.cfgs[name]
        return True

    def normalize(self):
        """ Update messages, commands and configs dictionary keys.  """
        normalize(self.msgs)
        normalize(self.cmds)
        normalize(self.cfgs)

        for key, msg in self.msgs.items():
            msg.normalize()

        for key, cmd in self.cmds.items():
            cmd.normalize()

        for key, cfg in self.cfgs.items():
            cfg.normalize()

    def get_cmd(self, name: Union[str, int]) -> Optional[Command]:
        if type(name) is str:
            return self.cmds.get(name)
        if type(name) is int or type(name) is float:
            for cmd in self.cmds.values():
                if cmd.id == int(name):
                    return cmd

    def __hash__(self):
        return hash((self.name, self.id, self.creation_date))

    def __repr__(self):
        return f"<Device name={self.name} id={self.id}>"




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'devicewidget.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_DeviceWidget(object):
    def setupUi(self, DeviceWidget):
        if not DeviceWidget.objectName():
            DeviceWidget.setObjectName(u"DeviceWidget")
        DeviceWidget.resize(370, 118)
        self.horizontalLayout_2 = QHBoxLayout(DeviceWidget)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.groupBox_2 = QGroupBox(DeviceWidget)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.horizontalLayout_3 = QHBoxLayout(self.groupBox_2)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.name = QLabel(self.groupBox_2)
        self.name.setObjectName(u"name")

        self.horizontalLayout_3.addWidget(self.name)

        self.id = QLabel(self.groupBox_2)
        self.id.setObjectName(u"id")

        self.horizontalLayout_3.addWidget(self.id)

        self.deleteButton = QPushButton(self.groupBox_2)
        self.deleteButton.setObjectName(u"deleteButton")
        self.deleteButton.setMaximumSize(QSize(25, 25))

        self.horizontalLayout_3.addWidget(self.deleteButton)

        self.deviceDetailsButton = QPushButton(self.groupBox_2)
        self.deviceDetailsButton.setObjectName(u"deviceDetailsButton")
        self.deviceDetailsButton.setMaximumSize(QSize(60, 25))
        self.deviceDetailsButton.setCheckable(True)

        self.horizontalLayout_3.addWidget(self.deviceDetailsButton)


        self.horizontalLayout_2.addWidget(self.groupBox_2)


        self.retranslateUi(DeviceWidget)

        QMetaObject.connectSlotsByName(DeviceWidget)
    # setupUi

    def retranslateUi(self, DeviceWidget):
        DeviceWidget.setWindowTitle(QCoreApplication.translate("DeviceWidget", u"DeviceWidget", None))
        self.groupBox_2.setTitle("")
        self.name.setText("")
        self.id.setText("")
        self.deleteButton.setText(QCoreApplication.translate("DeviceWidget", u"\u2717", None))
        self.deviceDetailsButton.setText(QCoreApplication.translate("DeviceWidget", u"...", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'cmddetails.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_CmdDetails(object):
    def setupUi(self, CmdDetails):
        if not CmdDetails.objectName():
            CmdDetails.setObjectName(u"CmdDetails")
        CmdDetails.resize(422, 429)
        CmdDetails.setMinimumSize(QSize(300, 0))
        self.horizontalLayout_5 = QHBoxLayout(CmdDetails)
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.argDetailsLayout = QVBoxLayout()
        self.argDetailsLayout.setObjectName(u"argDetailsLayout")
        self.groupBox = QGroupBox(CmdDetails)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout_4 = QVBoxLayout(self.groupBox)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.nameLabel = QLabel(self.groupBox)
        self.nameLabel.setObjectName(u"nameLabel")

        self.horizontalLayout.addWidget(self.nameLabel)

        self.nameEdit = QLineEdit(self.groupBox)
        self.nameEdit.setObjectName(u"nameEdit")

        self.horizontalLayout.addWidget(self.nameEdit)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.commentLabel = QLabel(self.groupBox)
        self.commentLabel.setObjectName(u"commentLabel")

        self.horizontalLayout_3.addWidget(self.commentLabel)

        self.commentEdit = QLineEdit(self.groupBox)
        self.commentEdit.setObjectName(u"commentEdit")

        self.horizontalLayout_3.addWidget(self.commentEdit)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.n_argsLabel = QLabel(self.groupBox)
        self.n_argsLabel.setObjectName(u"n_argsLabel")

        self.horizontalLayout_2.addWidget(self.n_argsLabel)

        self.n_argsEdit = QLineEdit(self.groupBox)
        self.n_argsEdit.setObjectName(u"n_argsEdit")

        self.horizontalLayout_2.addWidget(self.n_argsEdit)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.idLabel = QLabel(self.groupBox)
        self.idLabel.setObjectName(u"idLabel")

        self.horizontalLayout_4.addWidget(self.idLabel)

        self.idEdit = QLineEdit(self.groupBox)
        self.idEdit.setObjectName(u"idEdit")

        self.horizontalLayout_4.addWidget(self.idEdit)


        self.verticalLayout.addLayout(self.horizontalLayout_4)

        self.scrollArea = QScrollArea(self.groupBox)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self._CmdArgContents = QWidget()
        self._CmdArgContents.setObjectName(u"_CmdArgContents")
        self._CmdArgContents.setGeometry(QRect(0, 0, 374, 181))
        self.verticalLayout_3 = QVBoxLayout(self._CmdArgContents)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.CmdArgContents = QVBoxLayout()
        self.CmdArgContents.setObjectName(u"CmdArgContents")

        self.verticalLayout_3.addLayout(self.CmdArgContents)

        self.scrollArea.setWidget(self._CmdArgContents)

        self.verticalLayout.addWidget(self.scrollArea)

        self.addArgButton = QPushButton(self.groupBox)
        self.addArgButton.setObjectName(u"addArgButton")

        self.verticalLayout.addWidget(self.addArgButton)


        self.verticalLayout_4.addLayout(self.verticalLayout)

        self.deleteCmdButton = QPushButton(self.groupBox)
        self.deleteCmdButton.setObjectName(u"deleteCmdButton")

        self.verticalLayout_4.addWidget(self.deleteCmdButton)


        self.argDetailsLayout.addWidget(self.groupBox)


        self.horizontalLayout_5.addLayout(self.argDetailsLayout)


        self.retranslateUi(CmdDetails)

        QMetaObject.connectSlotsByName(CmdDetails)
    # setupUi

    def retranslateUi(self, CmdDetails):
        CmdDetails.setWindowTitle(QCoreApplication.translate("CmdDetails", u"CmdDetails", None))
        self.groupBox.setTitle(QCoreApplication.translate("CmdDetails", u"Command", None))
        self.nameLabel.setText(QCoreApplication.translate("CmdDetails", u"name", None))
        self.commentLabel.setText(QCoreApplication.translate("CmdDetails", u"comment", None))
        self.n_argsLabel.setText(QCoreApplication.translate("CmdDetails", u"n_args", None))
        self.idLabel.setText(QCoreApplication.translate("CmdDetails", u"id", None))
        self.addArgButton.setText(QCoreApplication.translate("CmdDetails", u"Add Arg", None))
        self.deleteCmdButton.setText(QCoreApplication.translate("CmdDetails", u"Delete", None))
    # retranslateUi




from .mainwindow import Ui_MainWindow
from .devicewidget import Ui_DeviceWidget
from .devicedetails import Ui_DeviceDetails
from .messagewidget import Ui_MessageWidget
from .messagedetails import Ui_MessageDetails
from .signalwidget import Ui_SignalWidget
from .signaldetails import Ui_SignalDetails
from .logwidget import Ui_LogWidget
from .enumwidget import Ui_EnumWidget
from .enumdetails import Ui_EnumDetails
from .logdetails import Ui_LogDetails
from .cfgwidget import Ui_CfgWidget
from .cfgdetails import Ui_CfgDetails
from .cmdwidget import Ui_CmdWidget
from .cmddetails import Ui_CmdDetails
from .cmdarg import Ui_CmdArg
from .enumvalue import Ui_EnumValue



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'notices.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_Notices(object):
    def setupUi(self, Notices):
        if not Notices.objectName():
            Notices.setObjectName(u"Notices")
        Notices.resize(400, 300)
        self.verticalLayout = QVBoxLayout(Notices)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.top_label = QLabel(Notices)
        self.top_label.setObjectName(u"top_label")

        self.verticalLayout.addWidget(self.top_label)

        self.label = QLabel(Notices)
        self.label.setObjectName(u"label")

        self.verticalLayout.addWidget(self.label)

        self.buttonBox = QDialogButtonBox(Notices)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(Notices)

        QMetaObject.connectSlotsByName(Notices)
    # setupUi

    def retranslateUi(self, Notices):
        Notices.setWindowTitle(QCoreApplication.translate("Notices", u"Form", None))
        self.top_label.setText("")
        self.label.setText("")
    # retranslateUi




from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .widgets.cmdwidget import Ui_CmdWidget
from .node_details import NodeDetails, FakeParent
from .cmd_details import CmdDetails

from ..specs import Command


class CmdWidget(QWidget):
    def __init__(self, gui, parent):
        QWidget.__init__(self)
        self.ui = Ui_CmdWidget()
        self.ui.setupUi(self)

        self.setVisible(False)

        self.parent = parent

        self.gui = gui

        self.details_button = None

        self.children = []

        self.ui.addCmdButton.clicked.connect(self.add_cmd)
        for cmd in self.parent.cmds.values():
            self.add_cmd(cmd)

    def save(self):
        for child in self.children:
            child.save()

        self.gui.reload()

    def reload(self):
        for child in self.children:
            child.reload()

    def add_cmd(self, cmd=None):
        if not cmd:
            cmd = Command(self.parent)

        self.parent.cmds[cmd.name] = cmd
        cmd_details = CmdDetails(self.gui, cmd, FakeParent())
        self.ui.CmdScrollContents.addWidget(cmd_details)
        self.children.append(cmd_details)

    def open_details(self, checked):
        for child in self.children:
            child.open_details(False)
            if not child.details_button is None:
                child.details_button.setChecked(False)



from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .widgets.messagedetails import Ui_MessageDetails
from .node_details import NodeDetails, FakeParent

from .signal_widget import SignalWidget
from .signal_details import SignalDetails

from ..specs import *


class MessageDetails(NodeDetails):
    def load_atts(self, ui, node):
        self.atts = [
            (ui.nameEdit, node.name, "name"),
            (ui.idEdit, node.id, "id"),
            (ui.dlcEdit, node.dlc, "dlc"),
            (ui.frequencyEdit, node.frequency, "frequency"),
            (ui.descriptionEdit, node.description, "description"),
        ]

    def __init__(self, gui, node: "Message", parent):
        NodeDetails.__init__(self, gui, node, parent)

        self.ui = Ui_MessageDetails()
        self.ui.setupUi(self)

        self.load_atts(self.ui, node)
        self.connect_atts()
 
        sigs = list(self.node.signals.values())
        for sig in sigs:
            self.add_node(sig)

        self.ui.deleteMessageButton.clicked.connect(self.delete)
        self.ui.addButton.clicked.connect(self.add_node)

        self.reload()
        self.setVisible(False)

    def add_node(self, node=None):
        if node == None or node == False:
            node = Signal(self.node)

        self.node.signals[node.name] = node
        s = SignalWidget(self.gui, node, SignalDetails, self.gui.ui.signalDetailsLayout)
        self.ui.verticalLayout_2.addWidget(s)
        self.children.append(s)



from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .widgets.devicedetails import Ui_DeviceDetails
from .node_details import NodeDetails, FakeParent
from .cfg_widget import CfgWidget
from .cmd_widget import CmdWidget
from .message_widget import MessageWidget
from .message_details import MessageDetails

from .undo_redo import UndoAdd

from ..specs import *


class DeviceDetails(NodeDetails):
    def load_atts(self, ui, node):
        self.atts = [
            (ui.idEdit, node.id, "id"),
            (ui.nameEdit, node.name, "name"),
        ]

    def __init__(self, gui, node: "Device", parent):
        NodeDetails.__init__(self, gui, node, parent)

        self.ui = Ui_DeviceDetails()
        self.ui.setupUi(self)

        self.load_atts(self.ui, self.node)
        self.connect_atts()

        self.children = []

        self.cfg_widget = CfgWidget(self.gui, self.node)
        self.children.append(self.cfg_widget)
        self.ui.cfgDetails.addWidget(self.cfg_widget)

        self.cmd_widget = CmdWidget(self.gui, self.node)
        self.children.append(self.cmd_widget)
        self.ui.cmdDetails.addWidget(self.cmd_widget)

        self.ui.addButton.clicked.connect(self.add_node)
        self.ui.cfgsButton.clicked.connect(self.open_cfg)
        self.ui.cmdsButton.clicked.connect(self.open_cmd)

        for msg in self.node.msgs.values():
            self.add_node(msg)

        self.ui.deleteDeviceButton.clicked.connect(self.delete)

        self.reload()

        self.setVisible(False)

    def open_cfg(self, clicked):
        self.cfg_widget.setVisible(clicked)
        return

    def open_cmd(self, clicked):
        self.cmd_widget.setVisible(clicked)
        return

    def add_node(self, node=None):
        if node == None or node == False:
            node = Message(self.node)

        self.node.msgs[node.name] = node
        m = MessageWidget(
            self.gui, node, MessageDetails, self.gui.ui.messageDetailsLayout
        )
        self.ui.verticalLayout.addWidget(m)
        self.children.append(m)


        undo_action = UndoAdd(node, m, m.delete, self.add_node)
        self.gui.undo_redo.push(undo_action)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'cmdwidget.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_CmdWidget(object):
    def setupUi(self, CmdWidget):
        if not CmdWidget.objectName():
            CmdWidget.setObjectName(u"CmdWidget")
        CmdWidget.resize(400, 300)
        CmdWidget.setMinimumSize(QSize(350, 0))
        self.verticalLayout = QVBoxLayout(CmdWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.groupBox = QGroupBox(CmdWidget)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout_4 = QVBoxLayout(self.groupBox)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.scrollArea = QScrollArea(self.groupBox)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scrollArea.setWidgetResizable(True)
        self._CmdScrollContents = QWidget()
        self._CmdScrollContents.setObjectName(u"_CmdScrollContents")
        self._CmdScrollContents.setGeometry(QRect(0, 0, 368, 208))
        self._CmdScrollContents.setMinimumSize(QSize(300, 0))
        self.verticalLayout_2 = QVBoxLayout(self._CmdScrollContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.CmdScrollContents = QVBoxLayout()
        self.CmdScrollContents.setObjectName(u"CmdScrollContents")

        self.verticalLayout_2.addLayout(self.CmdScrollContents)

        self.scrollArea.setWidget(self._CmdScrollContents)

        self.verticalLayout_4.addWidget(self.scrollArea)

        self.addCmdButton = QPushButton(self.groupBox)
        self.addCmdButton.setObjectName(u"addCmdButton")

        self.verticalLayout_4.addWidget(self.addCmdButton)


        self.verticalLayout.addWidget(self.groupBox)


        self.retranslateUi(CmdWidget)

        QMetaObject.connectSlotsByName(CmdWidget)
    # setupUi

    def retranslateUi(self, CmdWidget):
        CmdWidget.setWindowTitle(QCoreApplication.translate("CmdWidget", u"CmdWidget", None))
        self.groupBox.setTitle(QCoreApplication.translate("CmdWidget", u"Commands", None))
        self.addCmdButton.setText(QCoreApplication.translate("CmdWidget", u"Add Cmd", None))
    # retranslateUi




from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt


from .node_details import FakeParent
from .widgets.enumwidget import Ui_EnumWidget
from .enum_details import EnumDetails

from ..specs import Enum


class EnumWidget(QWidget):
    def __init__(self, gui, parent):
        self.Window = True
        QWidget.__init__(self)
        self.ui = Ui_EnumWidget()
        self.ui.setupUi(self)

        self.parent = parent
        self.gui = gui

        self.atts = []

        self.children = []

        self.reload()

        self.details = None

        self.ui.addEnumButton.clicked.connect(self.add_enum)
        for enum in self.parent.enums.values():
            self.add_enum(enum)

    def reload(self):
        for child in self.children:
            child.reload()

        for att, get_f, set_f in self.atts:
            att.setText(str(get_f()))

        return

    def save(self):
        for child in self.children:
            child.save()

        self.gui.reload()

    def add_enum(self, enum=None):
        if enum == None or enum == False:
            enum = Enum()
        self.parent.enums[enum.name] = enum
        enum_widget = EnumDetails(self.gui, enum, FakeParent())
        self.ui.enumContents.addWidget(enum_widget)
        self.children.append(enum_widget)



import sys
from pathlib import Path

from PySide2.QtWidgets import *

from .gui_widgets.mainwindow import MainWindow


def gui2(file: Path):
    app = QApplication([])
    window = MainWindow()

    if file is not None:
        window.load(file)

    window.show()
    sys.exit(app.exec_())



from typing import *
import datetime

from .node import Node
from .utils import normalize
from .enum_value import EnumValue

class Enum(Node):
    """Fcp Enum. C lookalike for FCP type definitions with name-value
    associations.
    """

    def __init__(self, parent: "Spec" = None) -> None:
        self.parent = parent
        self._name = ""
        self.enumeration = {}
        self.creation_date = datetime.datetime.now()

    def compile(self) -> Dict[str, Any]:
        enums = {k: v.compile() for (k, v) in self.enumeration.items()}

        d = make_public(self, (filter_private(self.__dict__)))
        d["enumeration"] = enums
        return d

    def decompile(self, d: Dict[str, Any]) -> None:
        """Transform node dictionary representation into a python class.

        :param d: Node dictionary
        """
        enumeration = d["enumeration"]
        del d["enumeration"]

        #self.__dict__.update(make_private(self, d))
        for k,v in self.make_private(self, d).items():
            self.__setattr__(k,v)

        for k, v in enumeration.items():
            enum_value = EnumValue(self)
            enum_value.decompile(v)
            self.enumeration[k] = enum_value

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self.name = name

    def normalize(self):
        normalize(self.enumeration)

    def __hash__(self):
        return hash((self.name, self.creation_date))

    def __repr__(self):
        return "name: {}".format(self.name)




import pytest

from ..fcp_lib import *
from ..can import CANMessage

@pytest.fixture
def fcp():
    with open("json/test.json") as f:
        j = json.loads(f.read())

    spec = Spec()
    spec.decompile(j)
    return Fcp(spec)

@pytest.fixture
def test_msg(fcp):
    return fcp.encode_msg("iib_motor", {"temp_motor":10})

def test_encode_msg(fcp, test_msg):
    msg = fcp.encode_msg("iib_motor", {"temp_motor":10})
    assert type(msg) is CANMessage
    assert msg.get_dev_id() == 16

def test_decode_msg(fcp, test_msg):
    name, signals = fcp.decode_msg(test_msg)
    assert name == "iib_motor"
    assert signals["temp_motor0"] == 10

def test_encode_cmd(fcp):
    msg = fcp.encode_cmd(10, "iib", "set_regen_on", [1,2,3])
    assert msg is not None

class FakeProxy():
    def __init__(self, socket, addrs):
        self.socket = socket
        self.addrs = []
        self.i = 0

    def recv(self) -> CANMessage:
        msgs = [
            CANMessage(sid=80, dlc=8, data16=[8<<8,2,0,0], timestamp=0),
            CANMessage(sid=208, dlc=8, data16=[1,1,0,0], timestamp=0),
            CANMessage(sid=144, dlc=8, data16=[1,1,0,0], timestamp=0)
        ]

        msg = msgs[self.i]
        self.i = (self.i + 1) % len(msgs)
        return Ok(msg)

    def send(self, msg: CANMessage):
        return

@pytest.fixture
def fcpcom(fcp):
    fcp_com = FcpCom(fcp, FakeProxy(None, None))
    return fcp_com

def test_fcpcom_cmd(fcpcom):
    fcpcom.start()
    rets = fcpcom.cmd("iib", "add", (1,1,0))
    fcpcom.stop()
    print(rets)
    assert rets.is_ok()
    assert rets.unwrap()[0] == 2

def test_fcpcom_set(fcpcom):
    fcpcom.start()
    rets = fcpcom.set("iib", "regen_on", 1)
    fcpcom.stop()
    print(rets)
    assert rets.is_ok()

def test_fcpcom_get(fcpcom):
    fcpcom.start()
    rets = fcpcom.get("iib", "regen_on")
    fcpcom.stop()
    print(rets)
    assert rets.is_ok()
    assert rets.unwrap() == 1

#def test_decode_msg(fcp):
#    msg = CANMessage(1232, 8, 0, data16=[0,0,0,0])
#    msg, signals = fcp.decode_msg(msg)
#    assert "motor_speed0" in signals.keys()
#    assert "n_motor_info" in signals.keys()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'enumvalue.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_EnumValue(object):
    def setupUi(self, EnumValue):
        if not EnumValue.objectName():
            EnumValue.setObjectName(u"EnumValue")
        EnumValue.resize(400, 85)
        self.verticalLayout = QVBoxLayout(EnumValue)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.groupBox = QGroupBox(EnumValue)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout_2 = QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.nameLabel = QLabel(self.groupBox)
        self.nameLabel.setObjectName(u"nameLabel")

        self.horizontalLayout_2.addWidget(self.nameLabel)

        self.nameEdit = QLineEdit(self.groupBox)
        self.nameEdit.setObjectName(u"nameEdit")

        self.horizontalLayout_2.addWidget(self.nameEdit)

        self.valueLabel = QLabel(self.groupBox)
        self.valueLabel.setObjectName(u"valueLabel")

        self.horizontalLayout_2.addWidget(self.valueLabel)

        self.valueEdit = QLineEdit(self.groupBox)
        self.valueEdit.setObjectName(u"valueEdit")

        self.horizontalLayout_2.addWidget(self.valueEdit)

        self.enumDeleteButton = QPushButton(self.groupBox)
        self.enumDeleteButton.setObjectName(u"enumDeleteButton")

        self.horizontalLayout_2.addWidget(self.enumDeleteButton)


        self.verticalLayout_2.addLayout(self.horizontalLayout_2)


        self.verticalLayout.addWidget(self.groupBox)


        self.retranslateUi(EnumValue)

        QMetaObject.connectSlotsByName(EnumValue)
    # setupUi

    def retranslateUi(self, EnumValue):
        EnumValue.setWindowTitle(QCoreApplication.translate("EnumValue", u"EnumDetails", None))
        self.groupBox.setTitle("")
        self.nameLabel.setText(QCoreApplication.translate("EnumValue", u"name", None))
        self.valueLabel.setText(QCoreApplication.translate("EnumValue", u"value", None))
        self.enumDeleteButton.setText(QCoreApplication.translate("EnumValue", u"Delete", None))
    # retranslateUi




from ..specs import *
import pytest

@pytest.fixture
def json():
    import json
    with open("json/test.json") as f:
        return json.loads(f.read())

@pytest.fixture
def spec(json):
    spec = Spec()
    spec.decompile(json)
    return spec



def test_make_sid():
    assert make_sid(10, 10) == 330

def test_decompose_id():
    assert decompose_id(330) == (10,10)

def test_decompile(json):
    spec = Spec()
    spec.decompile(json)

def test_compile(json):
    spec = Spec()
    spec.decompile(json)
    j = spec.compile()




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'cfgwidget.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_CfgWidget(object):
    def setupUi(self, CfgWidget):
        if not CfgWidget.objectName():
            CfgWidget.setObjectName(u"CfgWidget")
        CfgWidget.resize(400, 300)
        CfgWidget.setMinimumSize(QSize(300, 0))
        self.verticalLayout = QVBoxLayout(CfgWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.groupBox = QGroupBox(CfgWidget)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout_3 = QVBoxLayout(self.groupBox)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.scrollArea = QScrollArea(self.groupBox)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self._CfgScrollContents = QWidget()
        self._CfgScrollContents.setObjectName(u"_CfgScrollContents")
        self._CfgScrollContents.setGeometry(QRect(0, 0, 356, 209))
        self.verticalLayout_2 = QVBoxLayout(self._CfgScrollContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.CfgScrollContents = QVBoxLayout()
        self.CfgScrollContents.setObjectName(u"CfgScrollContents")

        self.verticalLayout_2.addLayout(self.CfgScrollContents)

        self.scrollArea.setWidget(self._CfgScrollContents)

        self.verticalLayout_3.addWidget(self.scrollArea)

        self.addCfgButton = QPushButton(self.groupBox)
        self.addCfgButton.setObjectName(u"addCfgButton")

        self.verticalLayout_3.addWidget(self.addCfgButton)


        self.verticalLayout.addWidget(self.groupBox)


        self.retranslateUi(CfgWidget)

        QMetaObject.connectSlotsByName(CfgWidget)
    # setupUi

    def retranslateUi(self, CfgWidget):
        CfgWidget.setWindowTitle(QCoreApplication.translate("CfgWidget", u"CfgWidget", None))
        self.groupBox.setTitle(QCoreApplication.translate("CfgWidget", u"Configs", None))
        self.addCfgButton.setText(QCoreApplication.translate("CfgWidget", u"Add Cfg", None))
    # retranslateUi




from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .widgets.cfgwidget import Ui_CfgWidget
from .cfg_details import CfgDetails
from .node_details import NodeDetails, FakeParent

from ..specs import Config


class CfgWidget(QWidget):
    def __init__(self, gui, parent):
        QWidget.__init__(self)
        self.ui = Ui_CfgWidget()
        self.ui.setupUi(self)

        self.setVisible(False)

        self.parent = parent

        self.gui = gui

        self.ui.addCfgButton.clicked.connect(self.add_cfg)

        self.children = []

        self.details_button = None

        for cfg in self.parent.cfgs.values():
            self.add_cfg(cfg)

    def open_details(self, checked):
        for child in self.children:
            child.open_details(False)
            if not child.details_button is None:
                child.details_button.setChecked(False)

    def save(self):
        for child in self.children:
            child.save()

        self.gui.reload()

    def reload(self):
        for child in self.children:
            child.reload()

    def add_cfg(self, cfg=None):
        if not cfg:
            cfg = Config()

        self.parent.cfgs[cfg.name] = cfg
        cfg_details = CfgDetails(self.gui, cfg, FakeParent())
        self.ui.CfgScrollContents.addWidget(cfg_details)
        self.children.append(cfg_details)



from PySide2.QtWidgets import *
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

from .widgets.cmddetails import Ui_CmdDetails
from .node_details import NodeDetails, FakeParent
from .arg_details import ArgDetails

from ..specs import Argument


class CmdDetails(NodeDetails):
    def load_atts(self, ui, node):
        self.atts = [
            (self.ui.nameEdit, node.name, "name"),
            (self.ui.n_argsEdit, node.n_args, "n_args"),
            (self.ui.commentEdit, node.comment, "comment"),
            (self.ui.idEdit, node.id, "id"),
        ]

    def __init__(self, gui, node, parent):
        NodeDetails.__init__(self, gui, node, parent)

        self.ui = Ui_CmdDetails()
        self.ui.setupUi(self)

        self.load_atts(self.ui, node)
        self.connect_atts()

        self.ui.deleteCmdButton.clicked.connect(self.delete)

        self.reload()
        self.setVisible(True)

        self.ui.addArgButton.clicked.connect(self.add_arg)

        for arg in node.args.values():
            self.add_arg(arg)

    def add_arg(self, arg=None):
        if arg == None or arg == False:
            arg = Argument(self.parent)

        self.node.args[arg.name] = arg
        arg_widget = ArgDetails(self.gui, arg, FakeParent())
        self.children.append(arg_widget)
        self.ui.CmdArgContents.addWidget(arg_widget)
